# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
# unit tests part
import unittest
from gmv.conf.utils.struct_parser import Compiler, CompilerError

class TestParser(unittest.TestCase):
    """ TestParser Class """
    
    def setUp(self): #pylint: disable=C0103
        pass
        
    def test_simple_list_test(self):
        """ a first simple test with space and indent, dedents to eat"""
        
        the_string = "         [ 'a',     1.435, 3 ]"
        
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, [ 'a', 1.435, 3])
    
    def test_negative_number_test(self):
        """ a negative number test """
        the_string = "         [ '-10.4',     1.435, 3 ]"
        
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, [ '-10.4', 1.435, 3])
    
    def test_imbricated_lists_test(self):
        """ multiple lists within lists """
        
        the_string = "[a,b, [1,2,3,4, [456,6,'absdef'], 234, 2.456 ], aqwe, done]"
                
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, ['a', 'b', [1, 2, 3, 4, [456, 6, 'absdef'], 234, 2.456 ]\
                                      , 'aqwe', 'done'])
  
    def test_list_without_bracket_test(self):
        """ simple list without bracket test """
        
        the_string = " 'a', b"
                
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, ['a', 'b'])
    
    def test_list_without_bracket_test_2(self): #pylint: disable=C0103
        """ list without bracket test with a list inside """
        the_string = " 'a', b, ['a thing', 2]"
                
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, ['a', 'b', ['a thing', 2] ])
        
    def test_list_error(self):
        """ list error """
        the_string = "  a ]"
        
        compiler = Compiler()
        
        try:
            compiler.compile_list(the_string)
        except CompilerError, err:
            self.assertEqual(err.message, 'Expression "  a ]" cannot be converted as a list.')
    
    def test_list_unicode_val(self):
        """ list unicode val """
        the_string = "[ u'[Gmail]/Чаты', 'z' ]".decode('utf-8')

        #to be in the same conditions as the conf object need to decode utf-8 as
        # it is done automatically with the os.open(...., 'uft-8')
        
        compiler = Compiler()
        
        compiler.compile_list(the_string)
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, [ u'[Gmail]/Чаты', 'z' ])
        
    def test_special_character_in_string(self):#pylint: disable=C0103
        """ simple list without bracket test """
        
        the_string = " 'a@', b"
                
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, ['a@','b'])
        
    def test_list_error_2(self):
        """ unsupported char @ """
        the_string = " a @"
        
        compiler = Compiler()
        
        try:
            compiler.compile_list(the_string)
        except CompilerError, err:
            self.assertEqual(err.message, 'Unsupported token (type: @, value : OP) (line=1,col=3).')
        
    def test_simple_dict(self):
        """ simple dict """
        
        the_string = "{'a':1, b:2 }"
                
        compiler = Compiler()
        
        the_result = compiler.compile_dict(the_string)
        
        self.assertEqual(the_result, {'a':1, 'b':2 })
        
    def test_dict_error(self):
        """ dict error """
        the_string = "{'a':1, b:2 "
                
        compiler = Compiler()
        
        try:
            compiler.compile_dict(the_string)
        except CompilerError, err:
            self.assertEqual(err.message, 'Expression "{\'a\':1, b:2 " cannot be converted as a dict.')
        
    def test_dict_with_list(self):
        """ dict with list """
        
        the_string = "{'a':1, b:[1,2,3,4,5] }"
                
        compiler = Compiler()
        
        the_result = compiler.compile_dict(the_string)
        
        self.assertEqual(the_result, {'a':1, 'b':[1, 2, 3, 4, 5]})
        
    def test_list_with_dict(self):
        """ list with dict """
        
        the_string = "['a',1,'b',{2:3,4:5} ]"
                
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, ['a', 1, 'b', { 2 : 3 , 4 : 5} ])
        
    def test_noquotes_dict(self):
        """ no quotes dict """
        
        the_string = "{ no12: a b , no10:a}"
                
        compiler = Compiler()
        
        the_result = compiler.compile_dict(the_string)
        
        self.assertEqual(the_result, { 'no12': 'a b' , 'no10':'a'})
        
    def test_everything(self):
        """ everything """
        
        the_string = "['a',1,'b',{2:3,4:[1,'hello', no quotes, [1,2,3,{1:2,3:4}]]} ]"
                
        compiler = Compiler()
        
        the_result = compiler.compile_list(the_string)
        
        self.assertEqual(the_result, ['a', 1, 'b', \
                                      {2 : 3, \
                                       4: [1, 'hello', 'no quotes', [1, 2, 3, {1:2, 3:4 }]]} ])
        
def tests():
    """ Global test method """
    #suite = unittest.TestLoader().loadTestsFromModule(struct_parser)
    suite = unittest.TestLoader().loadTestsFromTestCase(TestParser)
    unittest.TextTestRunner(verbosity=2).run(suite)
        
        
if __name__ == '__main__':
    tests()



# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import socket
import sys
import datetime
import os
import signal
import traceback

import argparse
import imaplib
import gmv.log_utils as log_utils
import gmv.gmvault_utils as gmvault_utils
import gmv.gmvault as gmvault
import gmv.gmvault_export as gmvault_export
import gmv.collections_utils as collections_utils

from gmv.cmdline_utils  import CmdLineParser
from gmv.credential_utils import CredentialHelper

GMVAULT_VERSION = gmvault_utils.GMVAULT_VERSION

GLOBAL_HELP_EPILOGUE = """Examples:

a) Get help for each of the individual commands

#> gmvault sync -h
#> gmvault restore --help
#> gmvault check -h
#> gmvault export -h

"""

REST_HELP_EPILOGUE = """Examples:

a) Complete restore of your gmail account (backed up in ~/gmvault-db) into anewfoo.bar@gmail.com 

#> gmvault restore -d ~/gmvault-db anewfoo.bar@gmail.com

b) Quick restore (restore only the last 2 months to make regular updates) of your gmail account into anewfoo.bar@gmail.com 

#> gmvault restore --type quick -d ~/gmvault-db foo.bar@gmail.com

c) Restart a restore after a previous error (Gmail can cut the connection if it is too long)

#> gmvault restore -d ~/gmvault-db anewfoo.bar@gmail.com --resume

d) Apply a label to all restored emails

#> gmvault restore --apply-label "20120422-gmvault" -d ~/gmvault-db anewfoo.bar@gmail.com
"""

SYNC_HELP_EPILOGUE = """Examples:

a) Full synchronisation with email and oauth login in ./gmvault-db

#> gmvault sync foo.bar@gmail.com

b) Quick daily synchronisation (only the last 2 months are scanned)

#> gmvault sync --type quick foo.bar@gmail.com

c) Resume Full synchronisation from where it failed to not go through your mailbox again

#> gmvault sync foo.bar@gmail.com --resume

d) Encrypt stored emails to save them safely anywhere

#> gmvault sync foo.bar@gmail.com --encrypt

d) Custom synchronisation with an IMAP request for advance users

#> gmvault sync --type custom --imap-req "Since 1-Nov-2011 Before 10-Nov-2011" foo.bar@gmail.com

e) Custom synchronisation with an Gmail request for advance users.
   Get all emails with label work and sent by foo.

#> gmvault sync --type custom --gmail-req "in:work from:foo" foo.bar@gmail.com

"""

EXPORT_HELP_EPILOGUE = """Warning: Experimental Functionality requiring more testing.

Examples:

a) Export default gmvault-db ($HOME/gmvault-db or %HOME$/gmvault-db) as a maildir mailbox.

#> gmvault export ~/my-mailbox-dir

b) Export a gmvault-db as a mbox mailbox (compliant with Thunderbird).

#> gmvault export -d /tmp/gmvault-db /tmp/a-mbox-dir

c) Export only a limited set of labels from the default gmvault-db as a mbox mailbox (compliant with Thunderbird).

#> gmvault export -l "label1" -l "TopLabel/LabelLev1" /tmp/a-mbox-dir

d) Use one of the export type dedicated to a specific tool (dovecot or offlineIMAP)

#> gmvault export -t dovecot /tmp/a-dovecot-dir
"""

LOG = log_utils.LoggerFactory.get_logger('gmv')

class NotSeenAction(argparse.Action): #pylint:disable=R0903,w0232
    """
       to differenciate between a seen and non seen command
    """
    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            setattr(namespace, self.dest, 'empty')
        else:
            setattr(namespace, self.dest, values)

class GMVaultLauncher(object):
    """
       GMVault launcher handling the command parsing
    """
    
    SYNC_TYPES    = ['full', 'quick', 'custom']
    RESTORE_TYPES = ['full', 'quick']
    CHECK_TYPES   = ['full']
    EXPORT_TYPES  = collections_utils.OrderedDict([
                     ('offlineimap', gmvault_export.OfflineIMAP),
                     ('dovecot', gmvault_export.Dovecot),
                     ('maildir', gmvault_export.OfflineIMAP),
                     ('mbox', gmvault_export.MBox)])
    EXPORT_TYPE_NAMES = ", ".join(EXPORT_TYPES)
    
    DEFAULT_GMVAULT_DB = "%s/gmvault-db" % (os.getenv("HOME", "."))
    
    def __init__(self):
        """ constructor """
        super(GMVaultLauncher, self).__init__()
        
    @gmvault_utils.memoized
    def _create_parser(self): #pylint: disable=R0915
        """
           Create the argument parser
           Return the created parser
        """
        parser = CmdLineParser()
        
        parser.epilogue = GLOBAL_HELP_EPILOGUE

        parser.add_argument("-v", '--version', action='version', version='Gmvault v%s' % (GMVAULT_VERSION))
        
        subparsers = parser.add_subparsers(title='subcommands', help='valid subcommands.')
         
        # A sync command
        sync_parser = subparsers.add_parser('sync', \
                                            help='synchronize with a given gmail account.')
        #email argument can be optional so it should be an option
        sync_parser.add_argument('email', \
                                 action='store', default='empty_$_email', help='email to sync with.')
        # sync typ
        sync_parser.add_argument('-t', '-type', '--type', \
                                 action='store', dest='type', \
                                 default='full', help='type of synchronisation: full|quick|custom. (default: full)')
        
        sync_parser.add_argument("-d", "--db-dir", \
                                 action='store', help="Database root directory. (default: $HOME/gmvault-db)",\
                                 dest="db_dir", default= self.DEFAULT_GMVAULT_DB)
               
        # for both when seen add const empty otherwise not_seen
        # this allow to distinguish between an empty value and a non seen option
        
        
        sync_parser.add_argument("-o", "--oauth", \
                          help="use oauth for authentication. (default recommended method)",\
                          action='store_const', dest="oauth_token", const='empty', default='not_seen')
        
        sync_parser.add_argument("-p", "--passwd", \
                          help="use interactive password authentication. (not recommended)",
                          action= 'store_const' , dest="passwd", const='empty', default='not_seen')
        
        sync_parser.add_argument("-2", "--2-legged-oauth", \
                          help="use 2 legged oauth for authentication. (Google Apps Business or Education accounts)",\
                          action='store_const', dest="two_legged_oauth_token", const='empty', default='not_seen')
        
        sync_parser.add_argument("--renew-oauth-tok", \
                          help="renew the stored oauth token (two legged or normal) via an interactive authentication session.",
                          action= 'store_const' , dest="oauth_token", const='renew')
         
        sync_parser.add_argument("--renew-passwd", \
                          help="renew the stored password via an interactive authentication session. (not recommended)",
                          action= 'store_const' , dest="passwd", const='renew')
        
        sync_parser.add_argument("--store-passwd", \
                          help="use interactive password authentication, encrypt and store the password. (not recommended)",
                          action= 'store_const' , dest="passwd", const='store')
        
        sync_parser.add_argument("-r", "--imap-req", metavar = "REQ", \
                                 help="Imap request to restrict sync.",\
                                 dest="imap_request", default=None)
        
        sync_parser.add_argument("-g", "--gmail-req", metavar = "REQ", \
                                 help="Gmail search request to restrict sync as defined in"\
                                      "https://support.google.com/mail/bin/answer.py?hl=en&answer=7190",\
                                 dest="gmail_request", default=None)
        
        # activate the resume mode --restart is deprecated
        sync_parser.add_argument("--resume", "--restart", \
                                 action='store_true', dest='restart', \
                                 default=False, help= 'Resume the sync action from the last saved gmail id.')
        
        # activate the resume mode --restart is deprecated
        sync_parser.add_argument("--emails-only", \
                                 action='store_true', dest='only_emails', \
                                 default=False, help= 'Only sync emails.')
        
        # activate the resume mode --restart is deprecated
        sync_parser.add_argument("--chats-only", \
                                 action='store_true', dest='only_chats', \
                                 default=False, help= 'Only sync chats.')
        
        sync_parser.add_argument("-e", "--encrypt", \
                                 help="encrypt stored email messages in the database.",\
                                 action='store_true',dest="encrypt", default=False)
        
        sync_parser.add_argument("-c", "--check-db", metavar = "VAL", \
                          help="enable/disable the removal from the gmvault db of the emails "\
                               "that have been deleted from the given gmail account. VAL = yes or no.",\
                          dest="db_cleaning", default=None)
        
        sync_parser.add_argument("-m", "--multiple-db-owner", \
                                 help="Allow the email database to be synchronized with emails from multiple accounts.",\
                                 action='store_true',dest="allow_mult_owners", default=False)
        
        # activate the restart mode
        sync_parser.add_argument("--no-compression", \
                                 action='store_false', dest='compression', \
                                 default=True, help= 'disable email storage compression (gzip).')
        
        sync_parser.add_argument("--server", metavar = "HOSTNAME", \
                              action='store', help="Gmail imap server hostname. (default: imap.gmail.com)",\
                              dest="host", default="imap.gmail.com")
            
        sync_parser.add_argument("--port", metavar = "PORT", \
                              action='store', help="Gmail imap server port. (default: 993)",\
                              dest="port", default=993)
        
        sync_parser.add_argument("--debug", "-debug", \
                              action='store_true', help="Activate debugging info",\
                              dest="debug", default=False)
        
        
        sync_parser.set_defaults(verb='sync')
    
        sync_parser.epilogue = SYNC_HELP_EPILOGUE
        
        # restore command
        rest_parser = subparsers.add_parser('restore', \
                                            help='restore gmvault-db to a given email account.')
        #email argument can be optional so it should be an option
        rest_parser.add_argument('email', \
                                 action='store', default='empty_$_email', help='email account to restore.')
        
        # restore typ
        rest_parser.add_argument('-t', '-type', '--type', \
                                 action='store', dest='type', \
                                 default='full', help='type of restoration: full|quick. (default: full)')
        
        # add a label
        rest_parser.add_argument('-a', '--apply-label', \
                                 action='store', dest='apply_label', \
                                 default=None, help='Apply a label to restored emails')
        
        # activate the resume mode --restart is deprecated
        rest_parser.add_argument("--resume", "--restart", \
                                 action='store_true', dest='restart', \
                                 default=False, help= 'Restart from the last saved gmail id.')
                                 
        # activate the resume mode --restart is deprecated
        rest_parser.add_argument("--emails-only", \
                                 action='store_true', dest='only_emails', \
                                 default=False, help= 'Only sync emails.')
        
        # activate the resume mode --restart is deprecated
        rest_parser.add_argument("--chats-only", \
                                 action='store_true', dest='only_chats', \
                                 default=False, help= 'Only sync chats.')
        
        rest_parser.add_argument("-d", "--db-dir", \
                                 action='store', help="Database root directory. (default: $HOME/gmvault-db)",\
                                 dest="db_dir", default= self.DEFAULT_GMVAULT_DB)
               
        # for both when seen add const empty otherwise not_seen
        # this allow to distinguish between an empty value and a non seen option
        rest_parser.add_argument("-o", "--oauth", \
                          help="use oauth for authentication. (default method)",\
                          action='store_const', dest="oauth_token", const='empty', default='not_seen')
        
        rest_parser.add_argument("-p", "--passwd", \
                          help="use interactive password authentication. (not recommended)",
                          action='store_const', dest="passwd", const='empty', default='not_seen')
        
        rest_parser.add_argument("-2", "--2-legged-oauth", \
                          help="use 2 legged oauth for authentication. (Google Apps Business or Education accounts)",\
                          action='store_const', dest="two_legged_oauth_token", const='empty', default='not_seen')
        
        
        rest_parser.add_argument("--server", metavar = "HOSTNAME", \
                              action='store', help="Gmail imap server hostname. (default: imap.gmail.com)",\
                              dest="host", default="imap.gmail.com")
            
        rest_parser.add_argument("--port", metavar = "PORT", \
                              action='store', help="Gmail imap server port. (default: 993)",\
                              dest="port", default=993)
        
        rest_parser.add_argument("--debug", "-debug", \
                              action='store_true', help="Activate debugging info",\
                              dest="debug", default=False)
        
        rest_parser.set_defaults(verb='restore')
    
        rest_parser.epilogue = REST_HELP_EPILOGUE
        
        # check_db command
        check_parser = subparsers.add_parser('check', \
                                            help='check and clean the gmvault-db disk database.')

        #email argument
        check_parser.add_argument('email', \
                                 action='store', default='empty_$_email', help='gmail account against which to check.')
        
        check_parser.add_argument("-d", "--db-dir", \
                                 action='store', help="Database root directory. (default: $HOME/gmvault-db)",\
                                 dest="db_dir", default= self.DEFAULT_GMVAULT_DB)
     
        # for both when seen add const empty otherwise not_seen
        # this allow to distinguish between an empty value and a non seen option
        check_parser.add_argument("-o", "--oauth", \
                          help="use oauth for authentication. (default method)",\
                          action='store_const', dest="oauth_token", const='empty', default='not_seen')
        
        check_parser.add_argument("-p", "--passwd", \
                          help="use interactive password authentication. (not recommended)",
                          action='store_const', dest="passwd", const='empty', default='not_seen')
        
        check_parser.add_argument("-2", "--2-legged-oauth", \
                          help="use 2 legged oauth for authentication. (Google Apps Business or Education accounts)",\
                          action='store_const', dest="two_legged_oauth_token", const='empty', default='not_seen')
        
        
        check_parser.add_argument("--server", metavar = "HOSTNAME", \
                              action='store', help="Gmail imap server hostname. (default: imap.gmail.com)",\
                              dest="host", default="imap.gmail.com")
            
        check_parser.add_argument("--port", metavar = "PORT", \
                              action='store', help="Gmail imap server port. (default: 993)",\
                              dest="port", default=993)
        
        check_parser.add_argument("--debug", "-debug", \
                              action='store_true', help="Activate debugging info",\
                              dest="debug", default=False)
        
        check_parser.set_defaults(verb='check')
        
        # export command
        export_parser = subparsers.add_parser('export', \
                                            help='Export the gmvault-db database to another format.')

        export_parser.add_argument('output_dir', \
                                   action='store', help='destination directory to export to.')

        export_parser.add_argument("-d", "--db-dir", \
                                 action='store', help="Database root directory. (default: $HOME/gmvault-db)",\
                                 dest="db_dir", default= self.DEFAULT_GMVAULT_DB)

        export_parser.add_argument('-t', '-type', '--type', \
                          action='store', dest='type', \
                          default='mbox', help='type of export: %s. (default: mbox)' % self.EXPORT_TYPE_NAMES)

        export_parser.add_argument('-l', '--label', \
                                   action='append', dest='label', \
                                   default=None,
                                   help='specify a label to export')
        export_parser.add_argument("--debug", "-debug", \
                       action='store_true', help="Activate debugging info",\
                       dest="debug", default=False)

        export_parser.set_defaults(verb='export')
        
        export_parser.epilogue = EXPORT_HELP_EPILOGUE

        return parser
      
    @classmethod
    def _parse_common_args(cls, options, parser, parsed_args, list_of_types = []): #pylint:disable=W0102
        """
           Parse the common arguments for sync and restore
        """
        #add email
        parsed_args['email']            = options.email
        
        parsed_args['debug']            = options.debug
        
        parsed_args['restart']          = options.restart
        
        #user entered both authentication methods
        if options.passwd == 'empty' and (options.oauth_token == 'empty' or options.two_legged_oauth_token == 'empty'):
            parser.error('You have to use one authentication method. '\
                         'Please choose between XOAuth and password (recommend XOAuth).')
        
        # user entered no authentication methods => go to default oauth
        if options.passwd == 'not_seen' and options.oauth_token == 'not_seen' and options.two_legged_oauth_token == 'not_seen':
            #default to xoauth
            options.oauth_token = 'empty'
            
        # add passwd
        parsed_args['passwd']           = options.passwd
        
        # add oauth tok
        if options.oauth_token == 'empty':
            parsed_args['oauth']      = options.oauth_token
            parsed_args['two_legged'] = False
        elif options.oauth_token == 'renew':
            parsed_args['oauth'] = 'renew'
            parsed_args['two_legged'] = True if options.two_legged_oauth_token == 'empty' else False          
        elif options.two_legged_oauth_token == 'empty':
            parsed_args['oauth']      = options.two_legged_oauth_token
            parsed_args['two_legged'] = True
        
        #add ops type
        if options.type:
            tempo_list = ['auto']
            tempo_list.extend(list_of_types)
            if options.type.lower() in tempo_list:
                parsed_args['type'] = options.type.lower()
            else:
                parser.error('Unknown type for command %s. The type should be one of %s' \
                             % (parsed_args['command'], list_of_types))
        
        #add db_dir
        parsed_args['db-dir']           = options.db_dir

        LOG.critical("Use gmvault-db located in %s.\n" % (parsed_args['db-dir'])) 
        
        # add host
        parsed_args['host']             = options.host
        
        #convert to int if necessary
        port_type = type(options.port)
        
        try:
            if port_type == type('s') or port_type == type("s"):
                port = int(options.port)
            else:
                port = options.port
        except Exception, _: #pylint:disable=W0703
            parser.error("--port option %s is not a number. Please check the port value" % (port))
            
        # add port
        parsed_args['port']             = port
             
        return parsed_args
    
    def parse_args(self): #pylint: disable=R0912
        """ Parse command line arguments 
            
            :returns: a dict that contains the arguments
               
            :except Exception Error
            
        """
        
        parser = self._create_parser()
          
        options = parser.parse_args()
        
        LOG.debug("Namespace = %s\n" % (options))
        
        parsed_args = { }
                
        parsed_args['command'] = options.verb
        
        if parsed_args.get('command', '') == 'sync':
            
            # parse common arguments for sync and restore
            self._parse_common_args(options, parser, parsed_args, self.SYNC_TYPES)
            
            # handle the search requests (IMAP or GMAIL dialect)
            if options.imap_request and options.gmail_request:
                parser.error('Please use only one search request type. You can use --imap-req or --gmail-req.')
            elif not options.imap_request and not options.gmail_request:
                LOG.debug("No search request type passed: Get everything.")
                parsed_args['request']   = {'type': 'imap', 'req':'ALL'}
            elif options.gmail_request and not options.imap_request:
                parsed_args['request']  = { 'type': 'gmail', 'req' : self._clean_imap_or_gm_request(options.gmail_request)}
            else:
                parsed_args['request']  = { 'type':'imap',  'req' : self._clean_imap_or_gm_request(options.imap_request)}
                
            # handle emails or chats only
            if options.only_emails and options.only_chats:
                parser.error("--emails-only and --chats-only cannot be used together. Please choose one.")
           
            parsed_args['emails_only'] = options.only_emails
            parsed_args['chats_only']  = options.only_chats
        
            # add db-cleaning
            # if request passed put it False unless it has been forced by the user
            # default is True (db-cleaning done)
            #default 
            parsed_args['db-cleaning'] = True
            
            # if there is a value then it is forced
            if options.db_cleaning: 
                parsed_args['db-cleaning'] = parser.convert_to_boolean(options.db_cleaning)
            
            #elif parsed_args['request']['req'] != 'ALL' and not options.db_cleaning:
            #    #else if we have a request and not forced put it to false
            #    parsed_args['db-cleaning'] = False
                
            if parsed_args['db-cleaning']:
                LOG.critical("Activate Gmvault db cleaning.")
            else:
                LOG.critical("Disable deletion of emails that are in Gmvault db and not anymore in Gmail.")
                
            #add encryption option
            parsed_args['encrypt'] = options.encrypt

            #add ownership checking
            parsed_args['ownership_control'] = not options.allow_mult_owners
            
            #compression flag
            parsed_args['compression'] = options.compression
                
                
        elif parsed_args.get('command', '') == 'restore':
            
            # parse common arguments for sync and restore
            self._parse_common_args(options, parser, parsed_args, self.RESTORE_TYPES)
            
            # apply restore labels if there is any
            parsed_args['apply_label'] = options.apply_label
            
            parsed_args['restart'] = options.restart
            
            # handle emails or chats only
            if options.only_emails and options.only_chats:
                parser.error("--emails-only and --chats-only cannot be used together. Please choose one.")
           
            parsed_args['emails_only'] = options.only_emails
            parsed_args['chats_only']  = options.only_chats
            
        elif parsed_args.get('command', '') == 'check':
            
            #add defaults for type
            options.type    = 'full'
            options.restart = False
            
            # parse common arguments for sync and restore
            self._parse_common_args(options, parser, parsed_args, self.CHECK_TYPES)
    
        elif parsed_args.get('command', '') == 'export':
            parsed_args['labels']     = options.label
            parsed_args['db-dir']     = options.db_dir
            parsed_args['output-dir'] = options.output_dir
            if options.type.lower() in self.EXPORT_TYPES:
                parsed_args['type'] = options.type.lower()
            else:
                parser.error('Unknown type for command export. The type should be one of %s' % self.EXPORT_TYPE_NAMES)
            parsed_args['debug'] = options.debug

        elif parsed_args.get('command', '') == 'config':
            pass
    
        #add parser
        parsed_args['parser']           = parser
        
        return parsed_args
    
    @classmethod
    def _clean_imap_or_gm_request(cls, request):
        """
           Clean request passed by the user with the option --imap-req or --gmail-req.
           Windows batch script preserve the single quote and unix shell doesn't.
           If the request starts and ends with single quote eat them.
        """
        LOG.debug("clean_imap_or_gm_request. original request = %s\n" % (request))
        
        if request and (len(request) > 2) and (request[0] == "'" and request[-1] == "'"):
            request =  request[1:-1]
            
        LOG.debug("clean_imap_or_gm_request. processed request = %s\n" % (request))
        return request
    
    @classmethod
    def _export(cls, args):
        export_type = cls.EXPORT_TYPES[args['type']]
        output_dir = export_type(args['output-dir'])
        LOG.critical("Export gmvault-db as a %s mailbox." % (args['type']))
        exporter = gmvault_export.GMVaultExporter(args['db-dir'], output_dir,
            labels=args['labels'])
        exporter.export()
        output_dir.close()

    @classmethod
    def _restore(cls, args, credential):
        """
           Execute All restore operations
        """
        LOG.critical("Connect to Gmail server.\n")
        # Create a gmvault releaving read_only_access
        restorer = gmvault.GMVaulter(args['db-dir'], args['host'], args['port'], \
                                       args['email'], credential, read_only_access = False)
        
        #full sync is the first one
        if args.get('type', '') == 'full':
            
            #call restore
            labels = [args['apply_label']] if args['apply_label'] else []
            restorer.restore(extra_labels = labels, restart = args['restart'], \
                             emails_only = args['emails_only'], chats_only = args['chats_only'])
            
        elif args.get('type', '') == 'quick':
            
            #take the last two to 3 months depending on the current date
            
            # today - 2 months
            today = datetime.date.today()
            begin = today - datetime.timedelta(gmvault_utils.get_conf_defaults().getint("Restore", "quick_days", 8))
            
            starting_dir = gmvault_utils.get_ym_from_datetime(begin)
            
            #call restore
            labels = [args['apply_label']] if args['apply_label'] else []
            restorer.restore(pivot_dir = starting_dir, extra_labels = labels, restart = args['restart'], \
                             emails_only = args['emails_only'], chats_only = args['chats_only'])
        
        else:
            raise ValueError("Unknown synchronisation mode %s. Please use full (default), quick.")
        
        #print error report
        LOG.critical(restorer.get_operation_report()) 
            
    @classmethod        
    def _sync(cls, args, credential):
        """
           Execute All synchronisation operations
        """
        LOG.critical("Connect to Gmail server.\n")
        
        # handle credential in all levels
        syncer = gmvault.GMVaulter(args['db-dir'], args['host'], args['port'], \
                                       args['email'], credential, read_only_access = True, use_encryption = args['encrypt'])
        
        
        
        #full sync is the first one
        if args.get('type', '') == 'full':
        
            #choose full sync. Ignore the request
            syncer.sync({ 'mode': 'full', 'type': 'imap', 'req': 'ALL' } , compress_on_disk = args['compression'], \
                        db_cleaning = args['db-cleaning'], ownership_checking = args['ownership_control'],\
                        restart = args['restart'], emails_only = args['emails_only'], chats_only = args['chats_only'])
        
        elif args.get('type', '') == 'auto':
        
            #choose auto sync. imap request = ALL and restart = True
            syncer.sync({ 'mode': 'auto', 'type': 'imap', 'req': 'ALL' } , compress_on_disk = args['compression'], \
                        db_cleaning = args['db-cleaning'], ownership_checking = args['ownership_control'],\
                        restart = True, emails_only = args['emails_only'], chats_only = args['chats_only'])
              
        elif args.get('type', '') == 'quick':
            
            #sync only the last x days (taken in defaults) in order to be quick 
            #(cleaning is import here because recent days might move again
            
            # today - 2 months
            today = datetime.date.today()
            begin = today - datetime.timedelta(gmvault_utils.get_conf_defaults().getint("Sync", "quick_days", 8))
            
            LOG.critical("Quick sync mode. Check for new emails since %s." % (begin.strftime('%d-%b-%Y')))
            
            # today + 1 day
            end   = today + datetime.timedelta(1)
            
            req   = { 'type' : 'imap', \
                      'req'  : syncer.get_imap_request_btw_2_dates(begin, end), \
                      'mode' : 'quick'}
            
            syncer.sync( req, \
                         compress_on_disk = args['compression'], \
                         db_cleaning = args['db-cleaning'], \
                         ownership_checking = args['ownership_control'], restart = args['restart'], \
                         emails_only = args['emails_only'], chats_only = args['chats_only'])
            
        elif args.get('type', '') == 'custom':
            
            #convert args to unicode
            args['request']['req']     = gmvault_utils.convert_to_unicode(args['request']['req'])
            args['request']['charset'] = 'utf-8' #for the moment always utf-8
            args['request']['mode']    = 'custom'

            # pass an imap request. Assume that the user know what to do here
            LOG.critical("Perform custom synchronisation with %s request: %s.\n" \
                         % (args['request']['type'], args['request']['req']))
            
            syncer.sync(args['request'], compress_on_disk = args['compression'], db_cleaning = args['db-cleaning'], \
                        ownership_checking = args['ownership_control'], restart = args['restart'], \
                        emails_only = args['emails_only'], chats_only = args['chats_only'])
        else:
            raise ValueError("Unknown synchronisation mode %s. Please use full (default), quick or custom.")
        
        
        #print error report
        LOG.critical(syncer.get_operation_report())
    
    @classmethod
    def _check_db(cls, args, credential):
        """
           Check DB
        """
        LOG.critical("Connect to Gmail server.\n")
        
        # handle credential in all levels
        checker = gmvault.GMVaulter(args['db-dir'], args['host'], args['port'], \
                                   args['email'], credential, read_only_access = True)
        
        checker.check_clean_db(db_cleaning = True)
            

    def run(self, args): #pylint:disable=R0912
        """
           Run the grep with the given args 
        """
        on_error       = True
        die_with_usage = True
        
        try:
            
            if args.get('command') not in ('export'):
                credential = CredentialHelper.get_credential(args)
            
            if args.get('command', '') == 'sync':
                
                self._sync(args, credential)
                
            elif args.get('command', '') == 'restore':
                
                self._restore(args, credential)
            
            elif args.get('command', '') == 'check':
                
                self._check_db(args, credential)
                
            elif args.get('command', '') == 'export':

                self._export(args)

            elif args.get('command', '') == 'config':
                
                LOG.critical("Configure something. TBD.\n")
            
            on_error = False
        
        except KeyboardInterrupt, _:
            LOG.critical("\nCRTL^C. Stop all operations.\n")
            on_error = False
        except socket.error:
            LOG.critical("Error: Network problem. Please check your gmail server hostname,"\
                         " the internet connection or your network setup.\n")
            LOG.critical("=== Exception traceback ===")
            LOG.critical(gmvault_utils.get_exception_traceback())
            LOG.critical("=== End of Exception traceback ===\n")
            die_with_usage = False
        except imaplib.IMAP4.error, imap_err:
            #bad login or password
            if str(imap_err) in ['[AUTHENTICATIONFAILED] Invalid credentials (Failure)', \
                                 '[ALERT] Web login required: http://support.google.com/'\
                                 'mail/bin/answer.py?answer=78754 (Failure)', \
                                 '[ALERT] Invalid credentials (Failure)'] :
                LOG.critical("ERROR: Invalid credentials, cannot login to the gmail server."\
                             " Please check your login and password or xoauth token.\n")
                die_with_usage = False
            else:
                LOG.critical("Error: %s. \n" % (imap_err) )
                LOG.critical("=== Exception traceback ===")
                LOG.critical(gmvault_utils.get_exception_traceback())
                LOG.critical("=== End of Exception traceback ===\n")
        except Exception, err:
            LOG.critical("Error: %s. \n" % (err) )
            LOG.critical("=== Exception traceback ===")
            LOG.critical(gmvault_utils.get_exception_traceback())
            LOG.critical("=== End of Exception traceback ===\n")
            die_with_usage = False
        finally: 
            if on_error and die_with_usage:
                args['parser'].die_with_usage()
 
def init_logging():
    """
       init logging infrastructure
    """       
    #setup application logs: one handler for stdout and one for a log file
    log_utils.LoggerFactory.setup_cli_app_handler(log_utils.STANDALONE, activate_log_file=False, file_path="./gmvault.log") 
    
def activate_debug_mode():
    """
       Activate debugging logging
    """
    LOG.critical("Debugging logs are going to be saved in file %s/gmvault.log.\n" % os.getenv("HOME","."))
    log_utils.LoggerFactory.setup_cli_app_handler(log_utils.STANDALONE, activate_log_file=True, \
                               console_level= 'DEBUG', file_path="%s/gmvault.log" % os.getenv("HOME","."))

def sigusr1_handler(signum, frame): #pylint:disable=W0613
    """
      Signal handler to get stack trace if the program is stuck
    """

    filename = './gmvault.traceback.txt'
    
    print("GMVAULT: Received SIGUSR1 -- Printing stack trace in %s..." % (os.path.abspath(filename)))

    the_f = open(filename, 'a')
    traceback.print_stack(file = the_f)
    the_f.close()

def register_traceback_signal():
    """ To register a USR1 signal allowing to get stack trace """
    signal.signal(signal.SIGUSR1, sigusr1_handler)

def setup_default_conf():
    """
       set the environment GMVAULT_CONF_FILE which is necessary for Conf object
    """
    gmvault_utils.get_conf_defaults() # force instanciation of conf to load the defaults

def bootstrap_run():
    """ temporary bootstrap """
    
    init_logging()
    
    #force argv[0] to gmvault
    sys.argv[0] = "gmvault"
    
    LOG.critical("")
    
    gmvlt = GMVaultLauncher()
    
    args = gmvlt.parse_args()
    
    #activate debug if enabled
    if args['debug']:
        LOG.critical("Activate debugging information.")
        activate_debug_mode()
    
    # force instanciation of conf to load the defaults
    gmvault_utils.get_conf_defaults() 
    
    gmvlt.run(args)
   
    
if __name__ == '__main__':
     
    #import memdebug
    
    #memdebug.start(8080)
    #import sys
    #print("sys.argv=[%s]" %(sys.argv))
    
    register_traceback_signal()
    
    bootstrap_run()
    
    sys.exit(0)



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import json
import gzip
import re
import os
import itertools
import fnmatch
import shutil

import gmv.blowfish as blowfish
import gmv.log_utils as log_utils

import gmv.collections_utils as collections_utils
import gmv.gmvault_utils as gmvault_utils
import gmv.imap_utils as imap_utils
import gmv.credential_utils as credential_utils

LOG = log_utils.LoggerFactory.get_logger('gmvault_db')
            
class GmailStorer(object): #pylint:disable=R0902,R0904,R0914
    '''
       Store emails on disk
    ''' 
    DATA_FNAME     = "%s/%s.eml"
    METADATA_FNAME = "%s/%s.meta"
    CHAT_GM_LABEL  = "gmvault-chats"
    
    ID_K         = 'gm_id'
    EMAIL_K      = 'email'
    THREAD_IDS_K = 'thread_ids'
    LABELS_K     = 'labels'
    INT_DATE_K   = 'internal_date'
    FLAGS_K      = 'flags'
    SUBJECT_K    = 'subject'
    MSGID_K      = 'msg_id'
    XGM_RECV_K   = 'x_gmail_received'
     
    HF_MSGID_PATTERN       = r"[M,m][E,e][S,s][S,s][a,A][G,g][E,e]-[I,i][D,d]:\s+<(?P<msgid>.*)>"
    HF_SUB_PATTERN         = r"[S,s][U,u][b,B][J,j][E,e][C,c][T,t]:\s+(?P<subject>.*)\s*"
    HF_XGMAIL_RECV_PATTERN = r"[X,x]-[G,g][M,m][A,a][I,i][L,l]-[R,r][E,e][C,c][E,e][I,i][V,v][E,e][D,d]:\s+(?P<received>.*)\s*"
    
    HF_MSGID_RE          = re.compile(HF_MSGID_PATTERN)
    HF_SUB_RE            = re.compile(HF_SUB_PATTERN)
    HF_XGMAIL_RECV_RE    = re.compile(HF_XGMAIL_RECV_PATTERN)
    
    ENCRYPTED_PATTERN = r"[\w+,\.]+crypt[\w,\.]*"
    ENCRYPTED_RE      = re.compile(ENCRYPTED_PATTERN)
    
    
    DB_AREA                    = 'db'
    QUARANTINE_AREA            = 'quarantine'
    CHATS_AREA                 = 'chats'
    BIN_AREA                   = 'bin'
    SUB_CHAT_AREA              = 'chats/%s'
    INFO_AREA                  = '.info'  # contains metadata concerning the database
    ENCRYPTION_KEY_FILENAME    = '.storage_key.sec'
    OLD_EMAIL_OWNER            = '.email_account.info' #deprecated
    EMAIL_OWNER                = '.owner_account.info'
    GMVAULTDB_VERSION          = '.gmvault_db_version.info'   
    
    def __init__(self, a_storage_dir, encrypt_data = False):
        """
           Store on disks
           args:
              a_storage_dir: Storage directory
              a_use_encryption: Encryption key. If there then encrypt
        """
        self._top_dir = a_storage_dir
        
        self._db_dir          = '%s/%s' % (a_storage_dir, GmailStorer.DB_AREA)
        self._quarantine_dir  = '%s/%s' % (a_storage_dir, GmailStorer.QUARANTINE_AREA)
        self._info_dir        = '%s/%s' % (a_storage_dir, GmailStorer.INFO_AREA)
        self._chats_dir       = '%s/%s' % (self._db_dir, GmailStorer.CHATS_AREA)
        self._bin_dir         = '%s/%s' % (a_storage_dir, GmailStorer.BIN_AREA)
        
        self._sub_chats_dir   = None
        self._sub_chats_inc   = -1
        self._sub_chats_nb    = -1
        
        self._limit_per_chat_dir = gmvault_utils.get_conf_defaults().getint("General", "limit_per_chat_dir", 1500)
        
        #make dirs
        if not os.path.exists(self._db_dir):
            LOG.critical("No Storage DB in %s. Create it.\n" % (a_storage_dir))
        
        gmvault_utils.makedirs(self._db_dir)
        gmvault_utils.makedirs(self._chats_dir)
        gmvault_utils.makedirs(self._quarantine_dir)
        gmvault_utils.makedirs(self._info_dir)
        
        self.fsystem_info_cache = {}
        
        self._encrypt_data   = encrypt_data
        self._encryption_key = None
        self._cipher         = None
        
        #add version if it is needed to migrate gmvault-db in the future
        self._create_gmvault_db_version()
        
        
    def _init_sub_chats_dir(self):
        """
           get info from existing sub chats
        """
        nb_to_dir = {}
        
        LOG.debug("LIMIT_PER_CHAT_DIR = %s" % (self._limit_per_chat_dir) )
        
        if os.path.exists(self._chats_dir):
            dirs = os.listdir(self._chats_dir)
            for the_dir in dirs:
                the_split = the_dir.split("-")
                if len(the_split) != 2:
                    raise Exception("Should get 2 elements in %s" % (the_split))
                
                nb_to_dir[int(the_split[1])] = the_dir
                
            
            if len(nb_to_dir) == 0:
                # no sub dir yet. Set it up
                self._sub_chats_nb  = 0
                self._sub_chats_inc = 1
                self._sub_chats_dir = self.SUB_CHAT_AREA % ("subchats-%s" % (self._sub_chats_inc))
                gmvault_utils.makedirs("%s/%s" % (self._db_dir, self._sub_chats_dir))
            
            # treat when more than limit chats in max dir 
            # treat when no dirs
            # add limit  as attribute limit_per_dir = 2000
            else:
                the_max = max(nb_to_dir)
                files = os.listdir("%s/%s" % (self._chats_dir, nb_to_dir[the_max]))
                self._sub_chats_nb  = len(files)/2
                self._sub_chats_inc = the_max
                self._sub_chats_dir = self.SUB_CHAT_AREA % nb_to_dir[the_max] 
            
        
    def get_sub_chats_dir(self):
        """
           Get sub_chats_dir
        """
        if self._sub_chats_inc == -1:
            self._init_sub_chats_dir()
         
        if self._sub_chats_nb >= self._limit_per_chat_dir:
            self._sub_chats_inc += 1
            
            self._sub_chats_nb  = 1 
            
            self._sub_chats_dir = self.SUB_CHAT_AREA % ("subchats-%s" % (self._sub_chats_inc))
            gmvault_utils.makedirs('%s/%s' % (self._db_dir, self._sub_chats_dir))
            
            return self._sub_chats_dir
        else:
            self._sub_chats_nb += 1
            return self._sub_chats_dir
    
    def _create_gmvault_db_version(self):
        """
           Create the Gmvault database version if it doesn't already exist
        """
        version_file = '%s/%s' % (self._info_dir, self.GMVAULTDB_VERSION)
        if not os.path.exists(version_file):
            the_fd = open(version_file, "w+")
            the_fd.write(gmvault_utils.GMVAULT_VERSION)
            the_fd.close()
    
    def store_db_owner(self, email_owner):
        """
           Store the email owner in .info dir. This is used to avoid synchronizing multiple email accounts in gmvault-db.
           Always wipe out completly the file
        """
        owners = self.get_db_owners()
        
        if email_owner not in owners:
            owners.append(email_owner)
            the_fd = open('%s/%s' % (self._info_dir, self.EMAIL_OWNER), "w+")
            json.dump(owners, the_fd, ensure_ascii = False)
            the_fd.flush()
            the_fd.close()
        
    
    def get_db_owners(self):
        """
           Get the email owner for the gmvault-db. Because except in particular cases, the db will be only linked to one meail.
        """
        fname = '%s/%s' % (self._info_dir, self.EMAIL_OWNER)
        if os.path.exists(fname):    
            the_fd = open(fname)
            list_of_owners = json.load(the_fd)
            the_fd.close()
            return list_of_owners
        
        return []
   
    def get_info_dir(self):
        """
           Return the info dir of gmvault-db
        """ 
        return self._info_dir
        
    def get_encryption_cipher(self):
        """
           Return the cipher to encrypt an decrypt.
           If the secret key doesn't exist, it will be generated.
        """
        if not self._cipher:
            if not self._encryption_key:
                self._encryption_key = credential_utils.CredentialHelper.get_secret_key('%s/%s' \
                % (self._info_dir, self.ENCRYPTION_KEY_FILENAME))
            
            #create blowfish cipher if data needs to be encrypted
            self._cipher = blowfish.Blowfish(self._encryption_key)
        
        return self._cipher
        
    @classmethod
    def get_encryption_key_path(cls, a_root_dir):
        """
           Return the path of the encryption key.
           This is used to print that information to the user
        """
        return  '%s/%s/%s' % (a_root_dir, cls.INFO_AREA, cls.ENCRYPTION_KEY_FILENAME)
    
    @classmethod
    def get_encryption_key(cls, a_info_dir):
        """
           Return or generate the encryption key if it doesn't exist
        """
        return credential_utils.CredentialHelper.get_secret_key('%s/%s' % (a_info_dir, cls.ENCRYPTION_KEY_FILENAME))
    
    @classmethod
    def parse_header_fields(cls, header_fields):
        """
           extract subject and message ids from the given header fields 
        """
        subject = None
        msgid   = None
        x_gmail_recv = None
        
        # look for subject
        matched = GmailStorer.HF_SUB_RE.search(header_fields)
        if matched:
            subject = matched.group('subject').strip()
        
        # look for a msg id
        matched = GmailStorer.HF_MSGID_RE.search(header_fields)
        if matched:
            msgid = matched.group('msgid').strip()

        # look for received xgmail id
        matched = GmailStorer.HF_XGMAIL_RECV_RE.search(header_fields)
        if matched:
            x_gmail_recv = matched.group('received').strip()
        
        return (subject, msgid, x_gmail_recv)
    
    def get_all_chats_gmail_ids(self):
        """
           Get only chats dirs 
        """
        # first create a normal dir and sort it below with an OrderedDict
        # beware orderedDict preserve order by insertion and not by key order
        gmail_ids = {}
        
        chat_dir = '%s/%s' % (self._db_dir, self.CHATS_AREA)
        if os.path.exists(chat_dir):
            the_iter = gmvault_utils.ordered_dirwalk(chat_dir, "*.meta")
        
            #get all ids
            for filepath in the_iter:
                directory, fname = os.path.split(filepath)
                gmail_ids[long(os.path.splitext(fname)[0])] = os.path.basename(directory)

            #sort by key 
            #used own orderedDict to be compliant with version 2.5
            gmail_ids = collections_utils.OrderedDict(sorted(gmail_ids.items(), key=lambda t: t[0]))
        
        return gmail_ids
        
        
    def get_all_existing_gmail_ids(self, pivot_dir = None, ignore_sub_dir = ['chats']): #pylint:disable=W0102
        """
           get all existing gmail_ids from the database within the passed month 
           and all posterior months
        """
        # first create a normal dir and sort it below with an OrderedDict
        # beware orderedDict preserve order by insertion and not by key order
        gmail_ids = {}
        
        if pivot_dir == None:
            #the_iter = gmvault_utils.dirwalk(self._db_dir, "*.meta")
            the_iter = gmvault_utils.ordered_dirwalk(self._db_dir, "*.meta", ignore_sub_dir)
        else:
            
            # get all yy-mm dirs to list
            dirs = gmvault_utils.get_all_dirs_posterior_to(pivot_dir, \
                   gmvault_utils.get_all_dirs_under(self._db_dir, ignore_sub_dir))
            
            #create all iterators and chain them to keep the same interface
            iter_dirs = [gmvault_utils.ordered_dirwalk('%s/%s' \
                        % (self._db_dir, the_dir), "*.meta", ignore_sub_dir) for the_dir in dirs]
            
            the_iter = itertools.chain.from_iterable(iter_dirs)
        
        #get all ids
        for filepath in the_iter:
            directory, fname = os.path.split(filepath)
            gmail_ids[long(os.path.splitext(fname)[0])] = os.path.basename(directory)

        #sort by key 
        #used own orderedDict to be compliant with version 2.5
        gmail_ids = collections_utils.OrderedDict(sorted(gmail_ids.items(), key=lambda t: t[0]))
        
        return gmail_ids
    
    def bury_chat_metadata(self, email_info, local_dir = None):
        """
           Like bury metadata but with an extra label gmvault-chat
        """
        extra_labels = [GmailStorer.CHAT_GM_LABEL]
        return self.bury_metadata(email_info, local_dir, extra_labels)
    
    def bury_metadata(self, email_info, local_dir = None, extra_labels = []): #pylint:disable=W0102
        """
            Store metadata info in .meta file
            Arguments:
             email_info: metadata info
             local_dir : intermdiary dir (month dir)
        """
        if local_dir:
            the_dir = '%s/%s' % (self._db_dir, local_dir)
            gmvault_utils.makedirs(the_dir)
        else:
            the_dir = self._db_dir
         
        meta_path = self.METADATA_FNAME % (the_dir, email_info[imap_utils.GIMAPFetcher.GMAIL_ID])
       
        meta_desc = open(meta_path, 'w')
        
        # parse header fields to extract subject and msgid
        subject, msgid, received = self.parse_header_fields(email_info[imap_utils.GIMAPFetcher.IMAP_HEADER_FIELDS_KEY])
        
        # need to convert labels that are number as string
        # come from imap_lib when label is a number
        labels = []
        for label in  email_info[imap_utils.GIMAPFetcher.GMAIL_LABELS]:
            if isinstance(label, (int, long, float, complex)):
                label = str(label)

            labels.append(unicode(gmvault_utils.remove_consecutive_spaces_and_strip(label)))
        
        labels.extend(extra_labels) #add extra labels
        
        #create json structure for metadata
        meta_obj = { 
                     self.ID_K         : email_info[imap_utils.GIMAPFetcher.GMAIL_ID],
                     self.LABELS_K     : labels,
                     self.FLAGS_K      : email_info[imap_utils.GIMAPFetcher.IMAP_FLAGS],
                     self.THREAD_IDS_K : email_info[imap_utils.GIMAPFetcher.GMAIL_THREAD_ID],
                     self.INT_DATE_K   : gmvault_utils.datetime2e(email_info[imap_utils.GIMAPFetcher.IMAP_INTERNALDATE]),
                     self.FLAGS_K      : email_info[imap_utils.GIMAPFetcher.IMAP_FLAGS],
                     self.SUBJECT_K    : subject,
                     self.MSGID_K      : msgid,
                     self.XGM_RECV_K   : received
                   }
        
        json.dump(meta_obj, meta_desc)
        
        meta_desc.flush()
        meta_desc.close()
         
        return email_info[imap_utils.GIMAPFetcher.GMAIL_ID]
    
    def bury_chat(self, chat_info, local_dir = None, compress = False):   
        """
            Like bury email but with a special label: gmvault-chats
            Arguments:
            chat_info: the chat content
            local_dir: intermediary dir
            compress : if compress is True, use gzip compression
        """
        extra_labels = ['gmvault-chats']
        
        return self.bury_email(chat_info, local_dir, compress, extra_labels)
    
    def bury_email(self, email_info, local_dir = None, compress = False, extra_labels = []): #pylint:disable=W0102
        """
           store all email info in 2 files (.meta and .eml files)
           Arguments:
             email_info: the email content
             local_dir : intermdiary dir (month dir)
             compress  : if compress is True, use gzip compression
        """
        
        if local_dir:
            the_dir = '%s/%s' % (self._db_dir, local_dir)
            gmvault_utils.makedirs(the_dir)
        else:
            the_dir = self._db_dir
        
        data_path = self.DATA_FNAME % (the_dir, email_info[imap_utils.GIMAPFetcher.GMAIL_ID])
        
        # if the data has to be encrypted
        if self._encrypt_data:
            data_path = '%s.crypt' % (data_path)
        
        if compress:
            data_path = '%s.gz' % (data_path)
            data_desc = gzip.open(data_path, 'wb')
        else:
            data_desc = open(data_path, 'wb')
            
        if self._encrypt_data:
            # need to be done for every encryption
            cipher = self.get_encryption_cipher()
            cipher.initCTR()
            data     = cipher.encryptCTR(email_info[imap_utils.GIMAPFetcher.EMAIL_BODY])
            gmvault_utils.buffered_write(data_desc, data) if len(data) > 4194304 else data_desc.write(data)
        else:
            
            data = email_info[imap_utils.GIMAPFetcher.EMAIL_BODY]
            #data_desc.write(data)
            gmvault_utils.buffered_write(data_desc, data) if len(data) > 4194304 else data_desc.write(data)
 
 
        self.bury_metadata(email_info, local_dir, extra_labels)
            
        data_desc.flush()
        data_desc.close()
        
        return email_info[imap_utils.GIMAPFetcher.GMAIL_ID]
    
    def get_directory_from_id(self, a_id, a_local_dir = None):
        """
           If a_local_dir (yy_mm dir) is passed, check that metadata file exists and return dir
           Return the directory path if id located.
           Return None if not found
        """
        filename = '%s.meta' % (a_id)
        
        #local_dir can be passed to avoid scanning the filesystem (because of WIN7 fs weaknesses)
        if a_local_dir:
            the_dir = '%s/%s' % (self._db_dir, a_local_dir)
            if os.path.exists(self.METADATA_FNAME % (the_dir, a_id)):
                return the_dir
            else:
                return None
        
        # first look in cache
        for the_dir in self.fsystem_info_cache:
            if filename in self.fsystem_info_cache[the_dir]:
                return the_dir
        
        #walk the filesystem
        for the_dir, _, files in os.walk(os.path.abspath(self._db_dir)):
            self.fsystem_info_cache[the_dir] = files
            for filename in fnmatch.filter(files, filename):
                return the_dir
        
        return None
    
    def _get_data_file_from_id(self, a_dir, a_id):
        """
           Return data file from the id
        """
        data_p = self.DATA_FNAME % (a_dir, a_id)
        
        # check if encrypted and compressed or not
        if os.path.exists('%s.crypt.gz' % (data_p)):
            data_fd = gzip.open('%s.crypt.gz' % (data_p), 'r')
        elif os.path.exists('%s.gz' % (data_p)):
            data_fd = gzip.open('%s.gz' % (data_p), 'r')
        elif os.path.exists('%s.crypt' % (data_p)):
            data_fd = open('%s.crypt' % (data_p), 'r')
        else:
            data_fd = open(data_p)
        
        return data_fd
    
    def _get_metadata_file_from_id(self, a_dir, a_id):
        """
           metadata file
        """
        meta_p = self.METADATA_FNAME % (a_dir, a_id)
       
        return open(meta_p)
    
    def quarantine_email(self, a_id):
        """
           Quarantine the email
        """
        #get the dir where the email is stored
        the_dir = self.get_directory_from_id(a_id)
        
        data = self.DATA_FNAME % (the_dir, a_id)
        
        # check if encrypted and compressed or not
        if os.path.exists('%s.crypt.gz' % (data)):
            data = '%s.crypt.gz' % (data)
        elif os.path.exists('%s.gz' % (data)):
            data = '%s.gz' % (data)
        elif os.path.exists('%s.crypt' % (data)):
            data = '%s.crypt' % (data)
        
        meta = self.METADATA_FNAME % (the_dir, a_id)

        #remove files if already quarantined
        q_data_path = os.path.join(self._quarantine_dir, os.path.basename(data))
        q_meta_path = os.path.join(self._quarantine_dir, os.path.basename(meta))

        if os.path.exists(q_data_path):
            os.remove(q_data_path)        
        
        if os.path.exists(q_meta_path):
            os.remove(q_meta_path)

        shutil.move(data, self._quarantine_dir)
        shutil.move(meta, self._quarantine_dir)
        
    def email_encrypted(self, a_email_fn):
        """
           True is filename contains .crypt otherwise False
        """
        basename = os.path.basename(a_email_fn)
        if self.ENCRYPTED_RE.match(basename):
            return True
        else:
            return False
        
    def unbury_email(self, a_id):
        """
           Restore the complete email info from info stored on disk
           Return a tuple (meta, data)
        """
        the_dir = self.get_directory_from_id(a_id)

        data_fd = self._get_data_file_from_id(the_dir, a_id)
        
        if self.email_encrypted(data_fd.name):
            LOG.debug("Restore encrypted email %s" % (a_id))
            # need to be done for every encryption
            cipher = self.get_encryption_cipher()
            cipher.initCTR()
            data = cipher.decryptCTR(data_fd.read())
        else:
            data = data_fd.read()
        
        return (self.unbury_metadata(a_id, the_dir), data)
    
    def unbury_data(self, a_id, a_id_dir = None):
        """
           Get the only the email content from the DB
        """
        if not a_id_dir:
            a_id_dir = self.get_directory_from_id(a_id)
            
        data_fd = self._get_data_file_from_id(a_id_dir, a_id)
        
        if self.email_encrypted(data_fd.name):
            LOG.debug("Restore encrypted email %s" % (a_id))
            # need to be done for every encryption
            cipher = self.get_encryption_cipher()
            cipher.initCTR()
            data = cipher.decryptCTR(data_fd.read())
        else:
            data = data_fd.read()
            
        return data    
        
    def unbury_metadata(self, a_id, a_id_dir = None):
        """
           Get metadata info from DB
        """
        if not a_id_dir:
            a_id_dir = self.get_directory_from_id(a_id)
        
        meta_fd = self._get_metadata_file_from_id(a_id_dir, a_id)
    
        metadata = json.load(meta_fd)
        
        metadata[self.INT_DATE_K] =  gmvault_utils.e2datetime(metadata[self.INT_DATE_K])
        
        # force convertion of labels as string because IMAPClient
        # returns a num when the label is a number (ie. '00000') and handle utf-8
        new_labels = []

        for label in  metadata[self.LABELS_K]:
            if isinstance(label, (int, long, float, complex)):
                label = str(label)
            new_labels.append(unicode(label))
 
        metadata[self.LABELS_K] = new_labels

        return metadata
    
    def delete_emails(self, emails_info, msg_type):
        """
           Delete all emails and metadata with ids
        """
        if msg_type == 'email':
            db_dir = self._db_dir
        else:
            db_dir = self._chats_dir

        move_to_bin = gmvault_utils.get_conf_defaults().get_boolean("General", "keep_in_bin" , False)

        if move_to_bin:
            LOG.critical("Move emails to the bin:%s" % (self._bin_dir))
        
        for (a_id, date_dir) in emails_info:
            
            the_dir = '%s/%s' % (db_dir, date_dir)
            
            data_p      = self.DATA_FNAME % (the_dir, a_id)
            comp_data_p = '%s.gz' % (data_p)
            cryp_comp_data_p = '%s.crypt.gz' % (data_p)
            
            metadata_p  = self.METADATA_FNAME % (the_dir, a_id)

            if move_to_bin:
                #move files to the bin
                gmvault_utils.makedirs(self._bin_dir)

                # create bin filenames
                bin_p          = self.DATA_FNAME % (self._bin_dir, a_id)
                metadata_bin_p = self.METADATA_FNAME % (self._bin_dir, a_id)
                
                if os.path.exists(data_p):
                    os.rename(data_p, bin_p)
                elif os.path.exists(comp_data_p):
                    os.rename(comp_data_p, '%s.gz' % (bin_p))
                elif os.path.exists(cryp_comp_data_p):
                    os.rename(cryp_comp_data_p, '%s.crypt.gz' % bin_p)   
                
                if os.path.exists(metadata_p):
                    os.rename(metadata_p, metadata_bin_p)
            else:
                #delete files if they exists
                if os.path.exists(data_p):
                    os.remove(data_p)
                elif os.path.exists(comp_data_p):
                    os.remove(comp_data_p)
                elif os.path.exists(cryp_comp_data_p):
                    os.remove(cryp_comp_data_p)   
                
                if os.path.exists(metadata_p):
                    os.remove(metadata_p)



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import sys
import os
import re
import codecs

import gmv.conf.exceptions as exceptions
import gmv.conf.utils.struct_parser as struct_parser                                      

class ResourceError(Exception):
    """
        Base class for ressource exceptions 
    """

    def __init__(self, a_msg):
        
        super(ResourceError, self).__init__(a_msg)

class Resource(object):
    """
        Class read a ressource.
        It can be read first from the Command Line, then from the ENV as an env variable and finally from a conf file 
    """
    
    def __init__(self, a_cli_argument=None, a_env_variable=None, a_conf_property=None): 
        """ 
          Default Constructor.
          It is important to understand that there is precedence between the different ways to set the ressource:
          - get from the command line if defined otherwise get from the Env variable if defined otherwise get from the conf file otherwise error
       
           Args:
              a_cli_argument : The command line argument name
              a_env_variable : The env variable name used for this ressource
              a_conf_property: It should be a tuple containing two elements (group,property)
        """
      
        self._cli_arg   = a_cli_argument.lower() if a_cli_argument is not None else None
        self._env_var   = a_env_variable.upper() if a_env_variable is not None else None
      
        if a_conf_property is not None:
            (self._conf_group, self._conf_property) = a_conf_property
        else:
            self._conf_group    = None
            self._conf_property = None
      
    def set_cli_argument(self, a_cli_argument):
        """cli_argument setter"""
        self._cli_arg = a_cli_argument.lower()
        
    def set_env_variable(self, a_env_variable):
        """env_variable setter"""
        self._env_var = a_env_variable
    
    @classmethod
    def _get_srandardized_cli_argument(cls, a_tostrip):
        """
           remove -- or - from the command line argument and add a -- prefix to standardize the cli argument 
        """
        the_str = a_tostrip
        
        while the_str.startswith('-'):
            the_str = the_str[1:]
        
        return '--%s' % (the_str)
    
    def _get_value_from_command_line(self):
        """
          internal method for extracting the value from the command line.
          All command line agruments must be lower case (unix style).
          To Do support short and long cli args.
           
           Returns:
             the Value if defined otherwise None
        """
          
        # check precondition
        if self._cli_arg == None:
            return None
        

        the_s = Resource._get_srandardized_cli_argument(self._cli_arg)
    
        # look for cliArg in sys argv
        for arg in sys.argv:
            if arg.lower() == the_s:
                i = sys.argv.index(arg)
                #print "i = %d, val = %s\n"%(i,sys.argv[i])
                if len(sys.argv) <= i:
                    # No more thing to read in the command line so quit
                    print "Resource: Commandline argument %s has no value\n" % (self._cli_arg)
                    return None 
                else:
                    #print "i+1 = %d, val = %s\n"%(i+1,sys.argv[i+1])
                    return sys.argv[i+1]
            

    def _get_value_from_env(self):
        """
          internal method for extracting the value from the env.
          All support ENV Variables should be in uppercase.
           
           Returns:
             the Value if defined otherwise None
        """
      
        # precondition
        if self._env_var == None:
            return None
     
        return os.environ.get(self._env_var, None)
      
    def _get_from_conf(self):
        """
           Try to read the info from the Configuration if possible
        """
        if self._conf_group and self._conf_property:
            if Conf.can_be_instanciated():
                return Conf.get_instance().get(self._conf_group, self._conf_property)
        
        return None
          
        
    def get_value(self, a_raise_exception=True):
        """
           Return the value of the Resource as a string.
           - get from the command line if defined otherwise get from the Env variable if defined otherwise get from the conf file otherwise error
              
           Arguments:
              aRaiseException: flag indicating if an exception should be raise if value not found
           Returns:
              value of the Resource as a String
       
           Raises:
              exception CTBTOError if the aRaiseExceptionOnError flag is activated
        """
       
        # get a value using precedence rule 1) command-line, 2) ENV, 3) Conf
        val = self._get_value_from_command_line()
        if val is None:
            val = self._get_value_from_env()
            if val is None:
                val = self._get_from_conf()
                if (val is None) and a_raise_exception:
                    
                    the_str = "Cannot find "
                    add_nor = 0
                    
                    if self._cli_arg is not None:
                        the_str += "commandline argument %s" % (self._cli_arg)
                        add_nor += 1
                    
                    if self._env_var is not None:
                        
                        if add_nor > 0:
                            the_str += ", nor "
                    
                        the_str += "the Env Variable %s" % (self._env_var)
                        add_nor += 1
                    
                    if self._conf_group is not None:
                        if add_nor > 0:
                            the_str += ", nor "
                        
                        the_str += "the Conf Group:[%s] and Property=%s" % (self._conf_group, self._conf_property)
                        add_nor += 1
                        
                    if add_nor == 0:
                        the_str += " any defined commandline argument, nor any env variable or"\
                                   " Conf group and properties. They are all None, fatal error"
                    else:
                        the_str += ". One of them should be defined"
                    
                    raise ResourceError(the_str)
    
        return val
   
    def _get(self, conv):
        """
           Private _get method used to convert to the right expected type (int,float or boolean).
           Strongly inspired by ConfigParser.py
              
           Returns:
              value converted into the asked type
       
           Raises:
              exception ValueError if conversion issue
        """
        return conv(self.get_value())

    def get_value_as_int(self):
        """
           Return the value as an int
              
           Returns:
              value converted into the asked type
       
           Raises:
              exception ValueError if conversion issue
        """
        return self._get(int)

    def get_value_as_float(self):
        """
           Return the value as a float
              
           Returns:
              value converted into the asked type
       
           Raises:
              exception ValueError if conversion issue
        """
        return self._get(float)

    _boolean_states = {'1': True, 'yes': True, 'true': True, 'on': True,
                       '0': False, 'no': False, 'false': False, 'off': False}

    def get_value_as_boolean(self):
        """
           Return the value as a boolean
              
           Returns:
              value converted into the asked type
       
           Raises:
              exception ValueError if conversion issue
        """
        val = self.get_value()
        if val.lower() not in self._boolean_states:
            raise ValueError, 'Not a boolean: %s' % val
        return self._boolean_states[val.lower()]

class MockConf(object):
    """
       MockConf Object that returns only defaults
    """
    def __init__(self, use_resource=True):
        """
           default constructor
        """
        pass
    
    @classmethod
    def get(cls, section, option, default=None, fail_if_missing=False): #pylint: disable=W0613
        """ get one option from a section.
        """
        return default
    
    @classmethod
    def print_content(cls, substitute_values = True):#pylint: disable=W0613
        """ print all the options variables substituted.
        
            :param a_substitue_vals: bool for substituting values
            :returns: the string containing all sections and variables
        """
        raise exceptions.Error("Not implemented in MockupConf")            

    @classmethod
    def items(cls, section):#pylint: disable=W0613
        """ return all items from a section. Items is a list of tuples (option,value)
            
            Args:
               section. The section where to find the option
               
            Returns: a list of tuples (option,value)
        
            Raises:
               exception NoSectionError if the section cannot be found
        """
        raise exceptions.Error("Not implemented in MockupConf") 
  
    @classmethod
    def getint(cls, section, option, default=0, fail_if_missing=False):#pylint: disable=W0613
        """Return the int value of the option.
        Default value is 0, None value can't be used as default value"""
        return default

    @classmethod
    def getfloat(cls, section, option, default=0, fail_if_missing=False):#pylint: disable=W0613
        """Return the float value of the option. 
        Default value is 0, None value can't be used as default value"""
        return default

    @classmethod
    def getboolean(cls, section, option, default=False, fail_if_missing=False):#pylint: disable=W0613
        """get bool value """
        return default
    
    @classmethod
    def get_list(cls, section, option, default=None, fail_if_missing=False):#pylint: disable=W0613
        """ get a list of string, int  """
        return default
    
    @classmethod
    def getlist(cls, section, option, default=None, fail_if_missing=False):#pylint: disable=W0613
        """ Deprecated, use get_list instead"""
        return cls.get_list(section, option, default, fail_if_missing)

    @classmethod
    def getdict(cls, section, option, default=None, fail_if_missing=False):#pylint: disable=W0613
        """ Deprecated, use get_dict instead"""
        return cls.get_dict(section, option, default, fail_if_missing)
        
    
    @classmethod
    def get_dict(cls, section, option, default=None, fail_if_missing=False):#pylint: disable=W0613
        """ get a dict """
        return default
 
class Conf(object):
    """ Configuration Object with a several features:
    
         * get configuration info in different types
         * support for import
         * support for variables in configuration file
         * support for default values in all accessors
         * integrated with the resources object offering to get the configuration from an env var, a commandline option or the conf
         * to be done : support for blocs, list comprehension and dict comprehension, json 
         * to be done : define resources in the conf using the [Resource] group with A= { ENV:TESTVAR, CLI:--testvar, VAL:1.234 }
    
    """
    # command line and env resource stuff
    CLINAME = "--conf_file"
    ENVNAME = "CONF_FILE" 
    
    #class member
    _instance = None
    
    _CLIGROUP = "CLI"
    _ENVGROUP = "ENV"
    _MAX_INCLUDE_DEPTH = 10
    
    @classmethod
    def get_instance(cls):
        """ singleton method """
        if cls._instance == None:
            cls._instance = Conf()
        return cls._instance
    
    @classmethod
    def can_be_instanciated(cls):
        """Class method used by the Resource to check that the Conf can be instantiated. 
        
        These two objects have a special contract as they are strongly coupled. 
        A Resource can use the Conf to check for a Resource and the Conf uses a Resource to read Conf filepath.
        
        :returns: True if the Conf file has got a file.
           
        :except Error: Base Conf Error
        
        """
        #No conf info passed to the resource so the Resource will not look into the conf (to avoid recursive search)
        the_res = Resource(cls.CLINAME, cls.ENVNAME)
        
        filepath = the_res.get_value(a_raise_exception=False)
        
        if (filepath is not None) and os.path.exists(filepath):
            return True
        
        return False
            
    
    def __init__(self, use_resource=True):
        """
           Constructor
        """
        
        # create resource for the conf file
        self._conf_resource = Resource(Conf.CLINAME, Conf.ENVNAME)
        
        # list of sections
        self._sections = {}
        
        self._configuration_file_path = None
        
        # create config object 
        if use_resource:       
            self._load_config()
        

   
    def _load_config(self, a_file = None):
        """ _load the configuration file """
        try:  
            # get it from a Resource if not files are passed
            if a_file is None:
                a_file = self._conf_resource.get_value() 
             
            if a_file is None:
                raise exceptions.Error("Conf. Error, need a configuration file path")
            
            #f_desc = open(a_file, 'r') 
            f_desc = codecs.open(a_file, 'r', 'utf-8') 
             
                
            self._read(f_desc, a_file)
            
            # memorize conf file path
            self._configuration_file_path = a_file
            
        except Exception, exce:
            print "Can't read the config file %s" % (a_file)
            print "Current executing from dir = %s\n" % (os.getcwd())
            raise exce
            
    
    def get_conf_file_path(self):
        """return conf_file_path"""
        return self._configuration_file_path if self._configuration_file_path != None else "unknown"
       
    def sections(self):
        """Return a list of section names, excluding [DEFAULT]"""
        # self._sections will never have [DEFAULT] in it
        return self._sections.keys()
    
    @classmethod
    def _get_defaults(cls, section, option, default, fail_if_missing):
        """ To manage defaults.
            Args:
               default. The default value to return if fail_if_missing is False
               fail_if_missing. Throw an exception when the option is not found and fail_if_missing is true
               
            Returns: default if fail_if_missing is False
        
            Raises:
               exception NoOptionError if fail_if_missing is True
        """
        if fail_if_missing:
            raise exceptions.Error(2, "No option %s in section %s" %(option, section))
        else:
            if default is not None:
                return str(default)
            else:
                return None
    
    def get(self, section, option, default=None, fail_if_missing=False):
        """ get one option from a section.
        
            return the default if it is not found and if fail_if_missing is False, otherwise return NoOptionError
          
            :param section: Section where to find the option
            :type  section: str
            :param option:  Option to get
            :param default: Default value to return if fail_if_missing is False
            :param fail_if_missing: Will throw an exception when the option is not found and fail_if_missing is true
               
            :returns: the option as a string
            
            :except NoOptionError: Raised only when fail_is_missing set to True
        
        """
        # all options are kept in lowercase
        opt = self.optionxform(option)
        
        if section not in self._sections:
            #check if it is a ENV section
            dummy = None
            if section == Conf._ENVGROUP:
                the_r = Resource(a_cli_argument=None, a_env_variable=opt)
                dummy = the_r.get_value()
            elif section == Conf._CLIGROUP:
                the_r = Resource(a_cli_argument=opt, a_env_variable=None)
                dummy = the_r.get_value()
            #return default if dummy is None otherwise return dummy
            return ((self._get_defaults(section, opt, default, fail_if_missing)) if dummy == None else dummy)
        elif opt in self._sections[section]:
            return self._replace_vars(self._sections[section][opt], "%s[%s]" % (section, option), - 1)
        else:
            return self._get_defaults(section, opt, default, fail_if_missing)
        
    
    def print_content(self, substitute_values = True):
        """ print all the options variables substituted.
        
            :param a_substitue_vals: bool for substituting values
            :returns: the string containing all sections and variables
        """
        
        result_str = ""
        
        for section_name in self._sections:
            result_str += "[%s]\n" % (section_name)
            section = self._sections[section_name]
            for option in section:
                if option != '__name__':
                    if substitute_values:
                        result_str += "%s = %s\n" % (option, self.get(section_name, option))
                    else:
                        result_str += "%s = %s\n" % (option, self._sections[section_name][option])
            
            result_str += "\n"
        
        return result_str
            

    def items(self, section):
        """ return all items from a section. Items is a list of tuples (option,value)
            
            Args:
               section. The section where to find the option
               
            Returns: a list of tuples (option,value)
        
            Raises:
               exception NoSectionError if the section cannot be found
        """
        try:
            all_sec = self._sections[section]
            # make a copy
            a_copy = all_sec.copy()
            # remove __name__ from d
            if "__name__" in a_copy:
                del a_copy["__name__"]
                
            return a_copy.items()
        
        except KeyError:
            raise exceptions.NoSectionError(section)

    def has_option(self, section, option):
        """Check for the existence of a given option in a given section."""
        has_option = False
        if self.has_section(section):
            option = self.optionxform(option)
            has_option = (option in self._sections[section])
        return has_option
    
    def has_section(self, section):
        """Check for the existence of a given section in the configuration."""
        has_section = False
        if section in self._sections:
            has_section = True
        return has_section
        
    @classmethod
    def _get_closing_bracket_index(cls, index, the_str, location, lineno):
        """ private method used by _replace_vars to count the closing brackets.
            
            Args:
               index. The index from where to look for a closing bracket
               s. The string to parse
               group. group and options that are substituted. Mainly used to create a nice exception message
               option. option that is substituted. Mainly used to create a nice exception message
               
            Returns: the index of the found closing bracket
        
            Raises:
               exception NoSectionError if the section cannot be found
        """
        
        tolook = the_str[index + 2:]
   
        opening_brack = 1
        closing_brack_index = index + 2
    
        i = 0
        for _ch in tolook:
            if _ch == ')':
                if opening_brack == 1:
                    return closing_brack_index
                else:
                    opening_brack -= 1
     
            elif _ch == '(':
                if tolook[i - 1] == '%':
                    opening_brack += 1
        
            # inc index
            closing_brack_index += 1
            i += 1
    
        raise exceptions.SubstitutionError(lineno, location, "Missing a closing bracket in %s" % (tolook))

    # very permissive regex
    _SUBSGROUPRE = re.compile(r"%\((?P<group>\w*)\[(?P<option>(.*))\]\)")
    
    def _replace_vars(self, a_str, location, lineno= - 1):
        """ private replacing all variables. A variable will be in the from of %(group[option]).
            Multiple variables are supported, ex /foo/%(group1[opt1])/%(group2[opt2])/bar
            Nested variables are also supported, ex /foo/%(group[%(group1[opt1]].
            Note that the group part cannot be substituted, only the option can. This is because of the Regular Expression _SUBSGROUPRE that accepts only words as values.
            
            Args:
               index. The index from where to look for a closing bracket
               s. The string to parse
               
            Returns: the final string with the replacements
        
            Raises:
               exception NoSectionError if the section cannot be found
        """
 
        toparse = a_str
    
        index = toparse.find("%(")
    
        # if found opening %( look for end bracket)
        if index >= 0:
            # look for closing brackets while counting openings one
            closing_brack_index = self._get_closing_bracket_index(index, a_str, location, lineno)
        
            #print "closing bracket %d"%(closing_brack_index)
            var   = toparse[index:closing_brack_index + 1]
            
            dummy = None
            
            matched = self._SUBSGROUPRE.match(var)
        
            if matched == None:
                raise exceptions.SubstitutionError(lineno, location, \
                                                   "Cannot match a group[option] in %s "\
                                                   "but found an opening bracket (. Malformated expression " \
                                                   % (var))
            else:
            
                # recursive calls
                group = self._replace_vars(matched.group('group'), location, - 1)
                option = self._replace_vars(matched.group('option'), location, - 1)
            
                try:
                    # if it is in ENVGROUP then check ENV variables with a Resource object
                    # if it is in CLIGROUP then check CLI argument with a Resource object
                    # otherwise check in standard groups
                    if group == Conf._ENVGROUP:
                        res = Resource(a_cli_argument=None, a_env_variable=option)
                        dummy = res.get_value()
                    elif group == Conf._CLIGROUP:
                        res = Resource(a_cli_argument=option, a_env_variable=None)
                        dummy = res.get_value()
                    else:
                        dummy = self._sections[group][self.optionxform(option)]
                except KeyError, _: #IGNORE:W0612
                    raise exceptions.SubstitutionError(lineno, location, "Property %s[%s] "\
                                                       "doesn't exist in this configuration file \n" \
                                                       % (group, option))
            
            toparse = toparse.replace(var, dummy)
            
            return self._replace_vars(toparse, location, - 1)    
        else:   
            return toparse 


    def _get(self, section, conv, option, default, fail_if_missing):
        """ Internal getter """
        return conv(self.get(section, option, default, fail_if_missing))

    def getint(self, section, option, default=0, fail_if_missing=False):
        """Return the int value of the option.
        Default value is 0, None value can't be used as default value"""
        return self._get(section, int, option, default, fail_if_missing)
    
    def get_int(self, section, option, default=0, fail_if_missing=False):
        """Return the int value of the option.
        Default value is 0, None value can't be used as default value"""
        return self._get(section, int, option, default, fail_if_missing)

    def getfloat(self, section, option, default=0, fail_if_missing=False):
        """Return the float value of the option. 
        Default value is 0, None value can't be used as default value"""
        return self._get(section, float, option, default, fail_if_missing)
    
    def get_float(self, section, option, default=0, fail_if_missing=False):
        """Return the float value of the option. 
        Default value is 0, None value can't be used as default value"""
        return self._get(section, float, option, default, fail_if_missing)

    _boolean_states = {'1': True, 'yes': True, 'true': True, 'on': True,
                       '0': False, 'no': False, 'false': False, 'off': False}

    def getboolean(self, section, option, default=False, fail_if_missing=False):
        """getboolean value""" 
        val = self.get(section, option, default, fail_if_missing)
        if val.lower() not in self._boolean_states:
            raise ValueError, 'Not a boolean: %s' % val
        return self._boolean_states[val.lower()]
    
    def get_boolean(self, section, option, default=False, fail_if_missing=False):
        """get_boolean value"""
        val = self.get(section, option, default, fail_if_missing)
        if val.lower() not in self._boolean_states:
            raise ValueError, 'Not a boolean: %s' % val
        return self._boolean_states[val.lower()]
    
    def get_list(self, section, option, default=None, fail_if_missing=False):
        """ get a list of string, int  """
        
        val = self.get(section, option, default, fail_if_missing)
        
        # parse it and return an error if invalid
        try:
            compiler = struct_parser.Compiler()
            return compiler.compile_list(val)
        except struct_parser.CompilerError, err: 
            raise exceptions.Error(err.message)
    
    def getlist(self, section, option, default=None, fail_if_missing=False):
        """ Deprecated, use get_list instead"""
        return self.get_list(section, option, default, fail_if_missing)

    def getdict(self, section, option, default=None, fail_if_missing=False):
        """ Deprecated, use get_dict instead"""
        return self.get_dict(section, option, default, fail_if_missing)
        
    
    def get_dict(self, section, option, default=None, fail_if_missing=False):
        """ get a dict """
        
        val = self.get(section, option, default, fail_if_missing)
        
        # parse it and return an error if invalid
        try:
            compiler = struct_parser.Compiler()
            return compiler.compile_dict(val)
        except struct_parser.CompilerError, err: 
            raise exceptions.Error(err.message)
        
    @classmethod
    def optionxform(cls, optionstr):
        """optionxform"""
        return optionstr.lower()
    
    #
    # Regular expressions for parsing section headers and options.
    #
    SECTCRE = re.compile(
        r'\['                                 # [
        r'(?P<header>[^]]+)'                  # very permissive!
        r'\]'                                 # ]
        )
    OPTCRE = re.compile(
        r'(?P<option>[^:=\s][^:=]*)'          # very permissive!
        r'\s*(?P<vi>[:=])\s*'                 # any number of space/tab,
                                              # followed by separator
                                              # (either : or =), followed
                                              # by any # space/tab
        r'(?P<value>.*)$'                     # everything up to eol
        )
            
    def _read_include(self, lineno, line, origin, depth):
        """_read_include"""      
        # Error if depth is MAX_INCLUDE_DEPTH 
        if depth >= Conf._MAX_INCLUDE_DEPTH:
            raise exceptions.IncludeError("Error. Cannot do more than %d nested includes."\
                                          " It is probably a mistake as you might have created a loop of includes" \
                                          % (Conf._MAX_INCLUDE_DEPTH))
        
        # remove %include from the path and we should have a path
        i = line.find('%include')
        
        #check if there is a < for including config files from a different format
        #position after include
        i = i + 8
        
        # include file with a specific reading module
        if line[i] == '<':
            dummy = line[i+1:].strip()
            f_i = dummy.find('>')
            if f_i == -1:
                raise exceptions.IncludeError("Error. > is missing in the include line no %s: %s."\
                                              " It should be %%include<mode:group_name> path" \
                                                   % (line, lineno), origin )
            else:
                group_name = None
                the_format     = dummy[:f_i].strip()
                
                the_list = the_format.split(':')
                if len(the_list) != 2 :
                    raise exceptions.IncludeError("Error. The mode and the group_name are not in the include line no %s: %s."\
                                                       " It should be %%include<mode:group_name> path" \
                                                       % (line, lineno), origin )
                else:
                    the_format, group_name = the_list
                    #strip the group name
                    group_name = group_name.strip()
                    
                path = dummy[f_i+1:].strip()
                
                # replace variables if there are any
                path = self._replace_vars(path, line, lineno)
                
                raise exceptions.IncludeError("External Module reading not enabled in this ConfHelper")
                #self._read_with_module(group_name, format, path, origin)
        else:
            # normal include   
            path = line[i:].strip() 
            
            # replace variables if there are any
            path = self._replace_vars(path, line, lineno)
            
            # check if file exits
            if not os.path.exists(path):
                raise exceptions.IncludeError("the config file to include %s does not exits" % (path), origin)
            else:
                # add include file and populate the section hash
                self._read(codecs.open(path, 'r', 'utf-8'), path, depth + 1)
                #self._read(open(path, 'r'), path, depth + 1)

    def _read(self, fpointer, fpname, depth=0): #pylint: disable=R0912
        """Parse a sectioned setup file.

        The sections in setup file contains a title line at the top,
        indicated by a name in square brackets (`[]'), plus key/value
        options lines, indicated by `name: value' format lines.
        Continuations are represented by an embedded newline then
        leading whitespace.  Blank lines, lines beginning with a '#',
        and just about everything else are ignored.
        Depth for avoiding looping in the includes
        """
        cursect = None                            # None, or a dictionary
        optname = None
        lineno = 0
        err = None                                  # None, or an exception
        while True:
            line = fpointer.readline()
            if not line:
                break
            lineno = lineno + 1
            # include in this form %include
            if line.startswith('%include'):
                self._read_include(lineno, line, fpname, depth)
                continue
            # comment or blank line?
            if line.strip() == '' or line[0] in '#;':
                continue
            if line.split(None, 1)[0].lower() == 'rem' and line[0] in "rR":
                # no leading whitespace
                continue
            # continuation line?
            if line[0].isspace() and cursect is not None and optname:
                value = line.strip()
                if value:
                    cursect[optname] = "%s\n%s" % (cursect[optname], value)
            # a section header or option header?
            else:
                # is it a section header?
                matched = self.SECTCRE.match(line)
                if matched:
                    sectname = matched.group('header')
                    if sectname in self._sections:
                        cursect = self._sections[sectname]
                    else:
                        cursect = {'__name__': sectname}
                        self._sections[sectname] = cursect
                    # So sections can't start with a continuation line
                    optname = None
                # no section header in the file?
                elif cursect is None:
                    raise exceptions.MissingSectionHeaderError(fpname, lineno, line)
                # an option line?
                else:
                    matched = self.OPTCRE.match(line)
                    if matched:
                        optname, vio, optval = matched.group('option', 'vi', 'value')
                        if vio in ('=', ':') and ';' in optval:
                            # ';' is a comment delimiter only if it follows
                            # a spacing character
                            pos = optval.find(';')
                            if pos != - 1 and optval[pos - 1].isspace():
                                optval = optval[:pos]
                        optval = optval.strip()
                        # allow empty values
                        if optval == '""':
                            optval = ''
                        optname = self.optionxform(optname.rstrip())
                        cursect[optname] = optval
                    else:
                        # a non-fatal parsing error occurred.  set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain a
                        # list of all bogus lines
                        if not err:
                            err = exceptions.ParsingError(fpname)
                        err.append(lineno, repr(line))
        # if any parsing errors occurred, raise an exception
        if err:
            raise err.get_error()



# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''

# Gmvault constants

GMAIL_UNLOCAL_CHATS = [
                     u'[Gmail]/Chats', u'[Google Mail]/Chats', #en, es, ger, portuguese
                     u'[Gmail]/Chat', u'[Google Mail]/Chat', #it
                     u'[Google Mail]/Tous les chats', u'[Gmail]/Tous les chats', # french
                     u'[Gmail]/Чаты', u'[Google Mail]/Чаты', # russian
                     u'[Google Mail]/Czat', u'[Gmail]/Czat', # polish
                     u'[Google Mail]/Bate-papos', u'[Gmail]/Bate-papos', #portuguese brazil
                    ]   # unlocalised Chats names

#The default conf file
DEFAULT_CONF_FILE = """#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  Gmvault Configuration file containing Gmvault defaults.
#  DO NOT CHANGE IT IF YOU ARE NOT AN ADVANCED USER
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

[Sync]
quick_days=10

[Restore]
# it is 10 days but currently it will always be the current month or the last 2 months
# the notion of days is not yet apparent in restore (only months).
quick_days=10
reserved_labels = [ u'migrated' ]

[General]
limit_per_chat_dir=2000
errors_if_chat_not_visible=False
nb_messages_per_batch=500
nb_messages_per_restore_batch=80
restore_default_location=DRAFTS
keep_in_bin=False

[Localisation]
#example with Russian
chat_folder=[ u'[Google Mail]/Чаты', u'[Gmail]/Чаты' ]
#uncomment if you need to force the term_encoding
#term_encoding='utf-8'

#Do not touch any parameters below as it could force an overwrite of this file
[VERSION]
conf_version=1.8-beta

#set environment variables for the program locally
#they will be read only once the conf file has been loaded
[ENV]
#by default it is ~/.gmvault
GMV_IMAP_DEBUG=0

"""



import time
import sys

def progress_2():
	"""
	"""
	percents = 0
	to_write = "Progress: [%s ]\r" % (percents)
	sys.stdout.write(to_write)
	sys.stdout.flush()

	steps = 100

	for i in xrange(steps):
		time.sleep(0.1)
		percents += 1
		#sys.stdout.write("\b" * (len(to_write)))
		to_write = "Progress: [%s percents]\r" % (percents)
		sys.stdout.write(to_write)
		sys.stdout.flush()



def progress_1():
	"""
	   progress_1
	"""
	toolbar_width = 100

	# setup toolbar
	sys.stdout.write("[%s]" % (" " * toolbar_width))
	sys.stdout.flush()
	sys.stdout.write("\b" * (toolbar_width+1)) # return to start of line, after '['

	for i in xrange(toolbar_width):
		time.sleep(0.1) # do real work here
		# update the bar
		sys.stdout.write("-")
		sys.stdout.flush()

	sys.stdout.write("\n")

if __name__ == '__main__':
	progress_2()



# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

Module containing the IMAPFetcher object which is the Wrapper around the modified IMAPClient object

'''
import math
import time
import socket
import re

import functools

import ssl
import imaplib

import gmv.gmvault_const as gmvault_const
import gmv.log_utils as log_utils
import gmv.credential_utils as credential_utils

import gmv.gmvault_utils as gmvault_utils
import gmv.mod_imap as mimap

LOG = log_utils.LoggerFactory.get_logger('imap_utils')

class PushEmailError(Exception):
    """
       PushEmail Error
    """
    def __init__(self, a_msg, quarantined = False):
        """
           Constructor
        """
        super(PushEmailError, self).__init__(a_msg)
        self._in_quarantine = quarantined
    
    def quarantined(self):
        """ Get email to quarantine """
        return self._in_quarantine

#retry decorator with nb of tries and sleep_time and backoff
def retry(a_nb_tries=3, a_sleep_time=1, a_backoff=1): #pylint:disable=R0912
    """
      Decorator for retrying command when it failed with a imap or socket error.
      Should be used exclusively on imap exchanges.
      Strategy, always retry on any imaplib or socket error. Wait few seconds before to retry
      backoff sets the factor by which the a_sleep_time should lengthen after each failure. backoff must be greater than 1,
      or else it isn't really a backoff
    """
    if a_backoff < 1:
        raise ValueError("a_backoff must be greater or equal to 1")

    a_nb_tries = math.floor(a_nb_tries)
    if a_nb_tries < 0:
        raise ValueError("a_nb_tries must be 0 or greater")

    if a_sleep_time <= 0:
        raise ValueError("a_sleep_time must be greater than 0")
    
    def reconnect(the_self, rec_nb_tries, total_nb_tries, rec_error, rec_sleep_time = [1]): #pylint: disable=W0102
        """
           Reconnect procedure. Sleep and try to reconnect
        """
        # go in retry mode if less than a_nb_tries
        while rec_nb_tries[0] < total_nb_tries:
            
            LOG.critical("Disconnecting from Gmail Server and sleeping ...")
            the_self.disconnect()            
            
            # add X sec of wait
            time.sleep(rec_sleep_time[0])
            rec_sleep_time[0] *= a_backoff #increase sleep time for next time
            
            rec_nb_tries[0] += 1
            
            #increase total nb of reconns
            the_self.total_nb_reconns += 1
           
            # go in retry mode: reconnect.
            # retry reconnect as long as we have tries left
            try:
                LOG.critical("Reconnecting to the from Gmail Server.")
                
                #reconnect to the current folder
                the_self.connect(go_to_current_folder = True )
                
                return 
            
            except Exception, ignored:
                # catch all errors and try as long as we have tries left
                LOG.exception(ignored)
        else:
            #cascade error
            raise rec_error
    
    def inner_retry(the_func): #pylint:disable=C0111,R0912
        def wrapper(*args, **kwargs): #pylint:disable=C0111,R0912
            nb_tries = [0] # make it mutable in reconnect
            m_sleep_time = [a_sleep_time]  #make it mutable in reconnect
            while True:
                try:
                    return the_func(*args, **kwargs)
                except PushEmailError, p_err:
                    
                    LOG.debug("error message = %s. traceback:%s" % (p_err, gmvault_utils.get_exception_traceback()))
                    
                    if nb_tries[0] < a_nb_tries:
                        LOG.critical("Cannot reach the Gmail server. Wait %s seconds and retrying." % (m_sleep_time[0]))
                    else:
                        LOG.critical("Stop retrying, tried too many times ...")
                    
                    reconnect(args[0], nb_tries, a_nb_tries, p_err, m_sleep_time)
                
                except imaplib.IMAP4.abort, err: #abort is recoverable and error is not
                    
                    LOG.debug("IMAP (abort) error message = %s. traceback:%s" % (err, gmvault_utils.get_exception_traceback()))
                    
                    if nb_tries[0] < a_nb_tries:
                        LOG.critical("Received an IMAP abort error. Wait %s seconds and retrying." % (m_sleep_time[0]))
                    else:
                        LOG.critical("Stop retrying, tried too many times ...")
                        
                    # problem with this email, put it in quarantine
                    reconnect(args[0], nb_tries, a_nb_tries, err, m_sleep_time)    
                    
                except socket.error, sock_err:
                    LOG.debug("error message = %s. traceback:%s" % (sock_err, gmvault_utils.get_exception_traceback()))
                    
                    if nb_tries[0] < a_nb_tries:
                        LOG.critical("Cannot reach the Gmail server. Wait %s seconds and retrying." % (m_sleep_time[0]))
                    else:
                        LOG.critical("Stop retrying, tried too many times ...")
                        
                    reconnect(args[0], nb_tries, a_nb_tries, sock_err, m_sleep_time)
                
                except ssl.SSLError, ssl_err:
                    LOG.debug("error message = %s. traceback:%s" % (ssl_err, gmvault_utils.get_exception_traceback()))
                    
                    if nb_tries[0] < a_nb_tries:
                        LOG.critical("Cannot reach the Gmail server. Wait %s seconds and retrying." % (m_sleep_time[0]))
                    else:
                        LOG.critical("Stop retrying, tried too many times ...")
                        
                    reconnect(args[0], nb_tries, a_nb_tries, sock_err, m_sleep_time)
                
                except imaplib.IMAP4.error, err:
                    
                    #just trace it back for the moment
                    LOG.debug("IMAP (normal) error message = %s. traceback:%s" % (err, gmvault_utils.get_exception_traceback()))
                    
                    if nb_tries[0] < a_nb_tries:
                        LOG.critical("Error when reaching Gmail server. Wait %s seconds and retry up to 2 times." \
                                     % (m_sleep_time[0]))
                    else:
                        LOG.critical("Stop retrying, tried too many times ...")
                    
                    #raise err
                    # retry 2 times before to quit
                    reconnect(args[0], nb_tries, 2, err, m_sleep_time)

        return functools.wraps(the_func)(wrapper)
        #return wrapper
    return inner_retry

class GIMAPFetcher(object): #pylint:disable=R0902,R0904
    '''
    IMAP Class reading the information
    '''
    GMAIL_EXTENSION     = 'X-GM-EXT-1'  # GMAIL capability
    GMAIL_ALL           = u'[Gmail]/All Mail' #GMAIL All Mail mailbox
    
    GENERIC_GMAIL_ALL   = u'\\AllMail' # unlocalised GMAIL ALL
    GENERIC_DRAFTS      = u'\\Drafts' # unlocalised DRAFTS
    GENERIC_GMAIL_CHATS = gmvault_const.GMAIL_UNLOCAL_CHATS   # unlocalised Chats names
    
    FOLDER_NAMES        = ['ALLMAIL', 'CHATS', 'DRAFTS']
    
    GMAIL_ID            = 'X-GM-MSGID' #GMAIL ID attribute
    GMAIL_THREAD_ID     = 'X-GM-THRID'
    GMAIL_LABELS        = 'X-GM-LABELS'
    
    IMAP_INTERNALDATE = 'INTERNALDATE'
    IMAP_FLAGS        = 'FLAGS'
    IMAP_ALL          = {'type':'imap', 'req':'ALL'}
    
    EMAIL_BODY        = 'BODY[]'
    
    GMAIL_SPECIAL_DIRS = ['\\Inbox', '\\Starred', '\\Sent', '\\Draft', '\\Important']
    
    #GMAIL_SPECIAL_DIRS_LOWER = ['\\inbox', '\\starred', '\\sent', '\\draft', '\\important']
    GMAIL_SPECIAL_DIRS_LOWER = ['\\inbox', '\\starred', '\\sent', '\\draft', '\\important', '\\trash']
    
    IMAP_BODY_PEEK     = 'BODY.PEEK[]' #get body without setting msg as seen

    #get the body info without setting msg as seen
    IMAP_HEADER_PEEK_FIELDS = 'BODY.PEEK[HEADER.FIELDS (MESSAGE-ID SUBJECT X-GMAIL-RECEIVED)]' 

    #key used to find these fields in the IMAP Response
    IMAP_HEADER_FIELDS_KEY      = 'BODY[HEADER.FIELDS (MESSAGE-ID SUBJECT X-GMAIL-RECEIVED)]'
    
    #GET_IM_UID_RE
    APPENDUID         = r'^[APPENDUID [0-9]* ([0-9]*)] \(Success\)$'
    
    APPENDUID_RE      = re.compile(APPENDUID)
    
    GET_ALL_INFO      = [ GMAIL_ID, GMAIL_THREAD_ID, GMAIL_LABELS, IMAP_INTERNALDATE, \
                          IMAP_BODY_PEEK, IMAP_FLAGS, IMAP_HEADER_PEEK_FIELDS]

    GET_ALL_BUT_DATA  = [ GMAIL_ID, GMAIL_THREAD_ID, GMAIL_LABELS, IMAP_INTERNALDATE, \
                          IMAP_FLAGS, IMAP_HEADER_PEEK_FIELDS]
    
    GET_DATA_ONLY     = [GMAIL_ID, IMAP_BODY_PEEK]
 
    GET_GMAIL_ID      = [ GMAIL_ID ]
    
    GET_GMAIL_ID_DATE = [ GMAIL_ID,  IMAP_INTERNALDATE]

    def __init__(self, host, port, login, credential, readonly_folder = True): #pylint:disable=R0913
        '''
            Constructor
        '''
        self.host                   = host
        self.port                   = port
        self.login                  = login
        self.once_connected         = False
        self.credential             = credential
        self.ssl                    = True
        self.use_uid                = True
        self.readonly_folder        = readonly_folder
        
        self.localized_folders      = { 'ALLMAIL': { 'loc_dir' : None, 'friendly_name' : 'allmail'}, 
                                        'CHATS'  : { 'loc_dir' : None, 'friendly_name' : 'chats'}, 
                                        'DRAFTS'  :{ 'loc_dir' : None, 'friendly_name' : 'drafts'} }
        
        # memoize the current folder (All Mail or Chats) for reconnection management
        self.current_folder        = None
        
        self.server                 = None
        self.go_to_all_folder       = True
        self.total_nb_reconns       = 0
        # True when CHATS or other folder error msg has been already printed
        self.printed_folder_error_msg = { 'ALLMAIL' : False, 'CHATS': False , 'DRAFTS':False }
        
        #update GENERIC_GMAIL_CHATS. Should be done at the class level
        self.GENERIC_GMAIL_CHATS.extend(gmvault_utils.get_conf_defaults().get_list('Localisation', 'chat_folder', []))
        
    def spawn_connection(self):
        """
           spawn a connection with the same parameters
        """
        conn = GIMAPFetcher(self.host, self.port, self.login, self.credential, self.readonly_folder)
        conn.connect()
        return conn
        
    def connect(self, go_to_current_folder = False):
        """
           connect to the IMAP server
        """
        # create imap object
        self.server = mimap.MonkeyIMAPClient(self.host, port = self.port, use_uid= self.use_uid, need_ssl= self.ssl)
        # connect with password or xoauth
        if self.credential['type'] == 'passwd':
            self.server.login(self.login, self.credential['value'])
        elif self.credential['type'] == 'xoauth':
            #connect with xoauth 
            if self.once_connected:
                #already connected once so renew xoauth req because it can expire
                self.credential['value'] = credential_utils.CredentialHelper.get_xoauth_req_from_email(self.login)
                
            self.server.xoauth_login(self.credential['value']) 
        else:
            raise Exception("Unknown authentication method %s. Please use xoauth or passwd authentication " \
                            % (self.credential['type']))
            
        #set connected to True to handle reconnection in case of failure
        self.once_connected = True
        
        # check gmailness
        self.check_gmailness()
         
        # find allmail chats and drafts folders
        self.find_folder_names()

        if go_to_current_folder and self.current_folder:
            self.server.select_folder(self.current_folder, readonly = self.readonly_folder)
            
        #enable compression
        self.enable_compression()
        LOG.debug("After Enabling compression")
            
    def disconnect(self):
        """
           disconnect to avoid too many simultaneous connection problem
        """
        if self.server:
            try:
                self.server.logout()
            except Exception, ignored: #ignored exception but still og it in log file if activated
                LOG.exception(ignored)
                
            self.server = None
    
    def reconnect(self):
        """
           disconnect and connect again
        """
        self.disconnect()
        self.connect()
    
    def enable_compression(self):
        """
           Try to enable the compression
        """
        self.server.enable_compression()
        
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def find_folder_names(self):
        """
           depending on your account the all mail folder can be named 
           [GMAIL]/ALL Mail or [GoogleMail]/All Mail.
           Find and set the right one
        """      
        #use xlist because of localized dir names
        folders = self.server.xlist_folders()
        
        the_dir = None
        for (flags, _, the_dir) in folders:
            #non localised GMAIL_ALL
            if GIMAPFetcher.GENERIC_GMAIL_ALL in flags:
                #it could be a localized Dir name
                self.localized_folders['ALLMAIL']['loc_dir'] = the_dir
            elif the_dir in GIMAPFetcher.GENERIC_GMAIL_CHATS :
                #it could be a localized Dir name
                self.localized_folders['CHATS']['loc_dir'] = the_dir
            elif GIMAPFetcher.GENERIC_DRAFTS in flags:
                self.localized_folders['DRAFTS']['loc_dir'] = the_dir
                
        if not self.localized_folders['ALLMAIL']['loc_dir']: # all mail error
            raise Exception("Cannot find global 'All Mail' folder (maybe localized and translated into your language) ! "\
                            "Check whether 'Show in IMAP for 'All Mail' is enabled in Gmail (Go to Settings->Labels->All Mail)")
        elif not self.localized_folders['CHATS']['loc_dir'] and \
                 gmvault_utils.get_conf_defaults().getboolean("General","errors_if_chat_not_visible", False):
            raise Exception("Cannot find global 'Chats' folder ! Check whether 'Show in IMAP for 'Chats' "\
                            "is enabled in Gmail (Go to Settings->Labels->All Mail)") 
        elif not self.localized_folders['DRAFTS']['loc_dir']:
            raise Exception("Cannot find global 'Drafts' folder.")
    
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def find_all_mail_folder(self):
        """
           depending on your account the all mail folder can be named 
           [GMAIL]/ALL Mail or [GoogleMail]/All Mail.
           Find and set the right one
        """      
        #use xlist because of localized dir names
        folders = self.server.xlist_folders()
        
        the_dir = None
        for (flags, _, the_dir) in folders:
            #non localised GMAIL_ALL
            if GIMAPFetcher.GENERIC_GMAIL_ALL in flags:
                #it could be a localized Dir name
                self.localized_folders['ALLMAIL']['loc_dir'] = the_dir
                return the_dir
        
        if not self.localized_folders['ALLMAIL']['loc_dir']:
            #Error
            raise Exception("Cannot find global 'All Mail' folder (maybe localized and translated into your language) !"\
                  " Check whether 'Show in IMAP for 'All Mail' is enabled in Gmail (Go to Settings->Labels->All Mail)")
        
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def find_chats_folder(self):
        """
           depending on your account the chats folder can be named 
           [GMAIL]/Chats or [GoogleMail]/Chats, [GMAIL]/tous les chats ...
           Find and set the right one
           Npte: Cannot use the flags as Chats is not a system label. Thanks Google
        """
        #use xlist because of localized dir names
        folders = self.server.xlist_folders()
        
        LOG.debug("Folders = %s\n" % (folders))
        
        the_dir = None
        for (_, _, the_dir) in folders:
            #look for GMAIL Chats
            if the_dir in GIMAPFetcher.GENERIC_GMAIL_CHATS :
                #it could be a localized Dir name
                self.localized_folders['CHATS']['loc_dir'] = the_dir
                return the_dir
        
        #Error did not find Chats dir 
        if gmvault_utils.get_conf_defaults().getboolean("General", "errors_if_chat_not_visible", False):
            raise Exception("Cannot find global 'Chats' folder ! Check whether 'Show in IMAP for 'Chats' "\
                            "is enabled in Gmail (Go to Settings->Labels->All Mail)") 
       
        return None
    
    def is_visible(self, a_folder_name):
        """
           check if a folder is visible otherwise 
        """
        dummy = self.localized_folders.get(a_folder_name)
        
        if dummy and (dummy.get('loc_dir', None) is not None):
            return True
            
        if not self.printed_folder_error_msg.get(a_folder_name, None): 
            LOG.critical("Cannot find 'Chats' folder on Gmail Server. If you wish to backup your chats,"\
                         " look at the documentation to see how to configure your Gmail account.\n")
            self.printed_folder_error_msg[a_folder_name] = True
        
          
        return False

    def get_folder_name(self, a_folder_name):
        """return real folder name from generic ones"""        
        if a_folder_name not in self.FOLDER_NAMES:
            raise Exception("%s is not a predefined folder names. Please use one" % (a_folder_name) )
            
        folder = self.localized_folders.get(a_folder_name, {'loc_dir' : 'GMVNONAME'})['loc_dir']

        return folder
           
    @retry(3,1,2)  # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def select_folder(self, a_folder_name, use_predef_names = True):
        """
           Select one of the existing folder
        """
        if use_predef_names:
            if a_folder_name not in self.FOLDER_NAMES:
                raise Exception("%s is not a predefined folder names. Please use one" % (a_folder_name) )
            
            folder = self.localized_folders.get(a_folder_name, {'loc_dir' : 'GMVNONAME'})['loc_dir']
            
            if self.current_folder != folder:
                self.server.select_folder(folder, readonly = self.readonly_folder)
                self.current_folder = folder
            
        elif self.current_folder != a_folder_name:
            self.server.select_folder(a_folder_name, readonly = self.readonly_folder)
            self.current_folder = a_folder_name
        
        return self.current_folder
        
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def list_all_folders(self): 
        """
           Return all folders mainly for debuging purposes
        """
        return self.server.xlist_folders()
        
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def get_capabilities(self):
        """
           return the server capabilities
        """
        if not self.server:
            raise Exception("GIMAPFetcher not connect to the GMAIL server")
        
        return self.server.capabilities()
    
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def check_gmailness(self):
        """
           Check that the server is a gmail server
        """
        if not GIMAPFetcher.GMAIL_EXTENSION in self.get_capabilities():
            raise Exception("GIMAPFetcher is not connected to a IMAP GMAIL server. Please check host (%s) and port (%s)" \
                  % (self.host, self.port))
        
        return True
    
    @retry(3,1,2) # try 3 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 4 sec
    def search(self, a_criteria):
        """
           Return all found ids corresponding to the search
        """
        return self.server.search(a_criteria)
    
    @retry(3,1,2) # try 4 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 8 sec
    def fetch(self, a_ids, a_attributes):
        """
           Return all attributes associated to each message
        """
        return self.server.fetch(a_ids, a_attributes)
                
    
    @classmethod
    def _old_build_labels_str(cls, a_labels):
        """
           Create IMAP label string from list of given labels. 
           Convert the labels to utf7
           a_labels: List of labels
        """
        # add GMAIL LABELS
        labels_str = None
        if a_labels and len(a_labels) > 0:
            labels_str = '('
            for label in a_labels:
                if gmvault_utils.contains_any(label, ' "*'):
                    label = label.replace('"', '\\"') #replace quote with escaped quotes
                    labels_str += '\"%s\" ' % (label)
                else:
                    labels_str += '%s ' % (label)
                    #labels_str += '\"%s\" ' % (label) #check if this is always ok or not
            
            labels_str = '%s%s' % (labels_str[:-1],')')
        
        return labels_str

    @classmethod
    def _build_labels_str(cls, a_labels):
        """
           Create IMAP label string from list of given labels. 
           Convert the labels to utf7
           a_labels: List of labels
        """
        # add GMAIL LABELS
        labels_str = None
        if a_labels and len(a_labels) > 0:
            labels_str = '('
            for label in a_labels:
                label = gmvault_utils.remove_consecutive_spaces_and_strip(label)
                #add not in self.GMAIL_SPECIAL_DIRS_LOWER
                if label.lower() in cls.GMAIL_SPECIAL_DIRS_LOWER:
                    labels_str += '%s ' % (label)
                else:
                    label = label.replace('"', '\\"') #replace quote with escaped quotes
                    labels_str += '\"%s\" ' % (label)
            labels_str = '%s%s' % (labels_str[:-1],')')
        
        return labels_str
    
    @classmethod
    def _get_dir_from_labels(cls, label):
        """
           Get the dirs to create from the labels
           
           label: label name with / in it
        """
        
        dirs = []
        
        i = 0
        for lab in label.split('/'):
            lab = gmvault_utils.remove_consecutive_spaces_and_strip(lab)
            if i == 0:
                dirs.append(lab)
            else:
                dirs.append('%s/%s' % (dirs[i-1], lab))
            
            i += 1
        
        return dirs
    
    def create_gmail_labels(self, labels, existing_folders):
        """
           Create folders and subfolders on Gmail in order
           to recreate the label hierarchy before to upload emails
           Note that adding labels with +X-GM-LABELS create only nested labels
           but not nested ones. This is why this trick must be used to 
           recreate the label hierarchy
           
           labels: list of labels to create
           
        """
        
        #1.5-beta moved that out of the loop to minimize the number of calls
        #to that method. (Could go further and memoize it)
        
        #get existing directories (or label parts)
        # get in lower case because Gmail labels are case insensitive
        listed_folders   = set([ directory.lower() for (_, _, directory) in self.list_all_folders() ])
        existing_folders = listed_folders.union(existing_folders)
        reserved_labels  = gmvault_utils.get_conf_defaults().get_list("Restore", "reserved_labels", \
                                                                      [ u'migrated' ])
        

        LOG.debug("Labels to create: [%s]" % (labels))
            
        for lab in labels:
            if lab.lower() in reserved_labels: #exclude creation of migrated label
                LOG.info("Warning ! label '%s' (lower or uppercase) is reserved by Gmail and cannot be used."\
                         "Use gmv-migrated instead" % (lab)) 
                lab = u"gmv-migrated"
           
            #split all labels
            labs = self._get_dir_from_labels(lab) 
            
            for directory in labs:
                low_directory = directory.lower() #get lower case directory but store original label
                if (low_directory not in existing_folders) and (low_directory not in self.GMAIL_SPECIAL_DIRS_LOWER):
                    try:
                        if self.server.create_folder(directory) != 'Success':
                            raise Exception("Cannot create label %s: the directory %s cannot be created." % (lab, directory))
                        else:
                            LOG.debug("============== ####### Created Labels (%s)." % (directory))
                    except imaplib.IMAP4.error, error:
                        #log error in log file if it exists
                        LOG.debug(gmvault_utils.get_exception_traceback())
                        if str(error).startswith("create failed: '[ALREADYEXISTS] Duplicate folder"):
                            LOG.critical("Warning: label %s already exists on Gmail and Gmvault tried to create it."\
                                         " Ignore this issue." % (directory) )
                        else:
                            raise error
                    
                    #add created folder in folders
                    existing_folders.add(low_directory)

        #return all existing folders
        return existing_folders
    
    
    @retry(3,1,2)
    def apply_labels_to(self, imap_ids, labels):
        """
           apply one labels to x emails
        """
        # go to All Mail folder
        LOG.debug("Applying labels %s" % (labels))
        
        the_timer = gmvault_utils.Timer()
        the_timer.start()

        #utf7 the labels as they should be
        labels = [ utf7_encode(label) for label in labels ]

        labels_str = self._build_labels_str(labels) # create labels str
    
        if labels_str:  
            #has labels so update email  
            the_timer.start()
            #LOG.debug("Before to store labels %s" % (labels_str))
            id_list = ",".join(map(str, imap_ids))
            #+X-GM-LABELS.SILENT to have not returned data
            ret_code, data = self.server._imap.uid('STORE', id_list, '+X-GM-LABELS.SILENT', labels_str) #pylint: disable=W0212

            #ret_code, data = self.server._imap.uid('COPY', id_list, labels[0])
            LOG.debug("After storing labels %s. Operation time = %s s.\nret = %s\ndata=%s" \
                      % (labels_str, the_timer.elapsed_ms(),ret_code, data))

            # check if it is ok otherwise exception
            if ret_code != 'OK':
                # Try again to code the error message (do not use .SILENT)
                ret_code, data = self.server._imap.uid('STORE', id_list, '+X-GM-LABELS', labels_str) #pylint: disable=W0212
                if ret_code != 'OK':
                    raise PushEmailError("Cannot add Labels %s to emails with uids %d. Error:%s" % (labels_str, imap_ids, data))
            else:
                LOG.debug("Stored Labels %s for gm_ids %s" % (labels_str, imap_ids))
       
    def delete_gmail_labels(self, labels, force_delete = False):
        """
           Delete passed labels. Beware experimental and labels must be ordered
        """
        for label in reversed(labels):
            
            labs = self._get_dir_from_labels(label)
            
            for directory in reversed(labs):
                
                if force_delete or ( (directory.lower() not in self.GMAIL_SPECIAL_DIRS_LOWER) \
                   and self.server.folder_exists(directory) ): #call server exists each time
                    try:
                        self.server.delete_folder(directory)
                    except imaplib.IMAP4.error, _:
                        LOG.debug(gmvault_utils.get_exception_traceback())
    
    
    def erase_mailbox(self):
        """
           This is for testing purpose and cannot be used with my own mailbox
        """
        
        if self.login == "guillaume.aubert@gmail.com":
            raise Exception("Error cannot activate erase_mailbox with %s" % (self.login))

        LOG.info("Erase mailbox for account %s." % (self.login))

        LOG.info("Delete folders")

        #delete folders
        folders = self.server.xlist_folders()

        LOG.debug("Folders = %s.\n" %(folders))

        trash_folder_name = None

        for (flags, _, the_dir) in folders:
            if (u'\\Starred' in flags) or (u'\\Spam' in flags) or (u'\\Sent' in flags) \
               or (u'\\Important' in flags) or (the_dir == u'[Google Mail]/Chats') \
               or (the_dir == u'[Google Mail]') or (u'\\Trash' in flags) or \
               (u'\\Inbox' in flags) or (GIMAPFetcher.GENERIC_GMAIL_ALL in flags) or \
               (GIMAPFetcher.GENERIC_DRAFTS in flags) or (GIMAPFetcher.GENERIC_GMAIL_CHATS in flags):
                LOG.info("Ignore folder %s" % (the_dir))           

                if (u'\\Trash' in flags): #keep trash folder name
                    trash_folder_name = the_dir
            else:
                LOG.info("Delete folder %s" % (the_dir))
                self.server.delete_folder(the_dir)
        
        self.select_folder('ALLMAIL')

        #self.server.store("1:*",'+X-GM-LABELS', '\\Trash')
        #self.server._imap.uid('STORE', id_list, '+X-GM-LABELS.SILENT', '\\Trash')
        #self.server.add_gmail_labels(self, messages, labels)

        LOG.info("Move emails to Trash.")
        
        # get all imap ids in ALLMAIL
        imap_ids = self.search(GIMAPFetcher.IMAP_ALL)

        #flag all message as deleted
        #print(self.server.delete_messages(imap_ids))

        if len(imap_ids) > 0:
            self.apply_labels_to(imap_ids, ['\\Trash'])

            LOG.info("Got all imap_ids flagged to Trash : %s." % (imap_ids))

        
        else:
            LOG.info("No messages to erase.")

        LOG.info("Delete emails from Trash.")

        if trash_folder_name == None:
            raise Exception("No trash folder ???")

        self.select_folder(trash_folder_name, False)
        
        # get all imap ids in ALLMAIL
        imap_ids = self.search(GIMAPFetcher.IMAP_ALL)

        if len(imap_ids) > 0:
            res = self.server.delete_messages(imap_ids)
            LOG.debug("Delete messages result = %s" % (res))

        LOG.info("Expunge everything.")
        self.server.expunge()

    @retry(4,1,2) # try 4 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 8 sec    
    def push_data(self, a_folder, a_body, a_flags, a_internal_time):
        """
           Push the data
        """  
        # protection against myself
        if self.login == 'guillaume.aubert@gmail.com':
            raise Exception("Cannot push to this account")
        
        the_timer = gmvault_utils.Timer()
        the_timer.start()
        LOG.debug("Before to Append email contents")
        #import sys  #to print the msg in stdout
        #import codecs
        #sys.stdout = codecs.getwriter('utf-8')(sys.__stdout__) 
        #msg = "a_folder = %s, a_flags = %s" % (a_folder.encode('utf-8'), a_flags)
        #msg = "a_folder = %s" % (a_folder.encode('utf-8'))
        #msg = msg.encode('utf-8')
        #print(msg)
        res = self.server.append(a_folder, a_body, a_flags, a_internal_time)
    
        LOG.debug("Appended data with flags %s and internal time %s. Operation time = %s.\nres = %s\n" \
                  % (a_flags, a_internal_time, the_timer.elapsed_ms(), res))
        
        # check res otherwise Exception
        if '(Success)' not in res:
            raise PushEmailError("GIMAPFetcher cannot restore email in %s account." %(self.login))
        
        match = GIMAPFetcher.APPENDUID_RE.match(res)
        if match:
            result_uid = int(match.group(1))
            LOG.debug("result_uid = %s" %(result_uid))
        else:
            # do not quarantine it because it seems to be done by Google Mail to forbid data uploading.
            raise PushEmailError("No email id returned by IMAP APPEND command. Quarantine this email.", quarantined = True)
        
        return result_uid          
         
    @retry(4,1,2) # try 4 times to reconnect with a sleep time of 1 sec and a backoff of 2. The fourth time will wait 8 sec
    def push_email(self, a_body, a_flags, a_internal_time, a_labels):
        """
           Push a complete email body 
        """
        #protection against myself
        if self.login == 'guillaume.aubert@gmail.com':
            raise Exception("Cannot push to this account")
    
        the_t = gmvault_utils.Timer()
        the_t.start()
        LOG.debug("Before to Append email contents")
        #res = self.server.append(self.current_folder, a_body, a_flags, a_internal_time)
        res = self.server.append(u'[Google Mail]/All Mail', a_body, a_flags, a_internal_time)
    
        LOG.debug("Appended data with flags %s and internal time %s. Operation time = %s.\nres = %s\n" \
                  % (a_flags, a_internal_time, the_t.elapsed_ms(), res))
        
        # check res otherwise Exception
        if '(Success)' not in res:
            raise PushEmailError("GIMAPFetcher cannot restore email in %s account." %(self.login))
        
        match = GIMAPFetcher.APPENDUID_RE.match(res)
        if match:
            result_uid = int(match.group(1))
            LOG.debug("result_uid = %s" %(result_uid))
        else:
            # do not quarantine it because it seems to be done by Google Mail to forbid data uploading.
            raise PushEmailError("No email id returned by IMAP APPEND command. Quarantine this email.", quarantined = True)
        
        labels_str = self._build_labels_str(a_labels)
        
        if labels_str:  
            #has labels so update email  
            the_t.start()
            LOG.debug("Before to store labels %s" % (labels_str))
            self.server.select_folder(u'[Google Mail]/All Mail', readonly = self.readonly_folder) # go to current folder
            LOG.debug("Changing folders. elapsed %s s\n" % (the_t.elapsed_ms()))
            the_t.start()
            ret_code, data = self.server._imap.uid('STORE', result_uid, '+X-GM-LABELS', labels_str) #pylint: disable=W0212
            #ret_code = self.server._store('+X-GM-LABELS', [result_uid],labels_str)
            LOG.debug("After storing labels %s. Operation time = %s s.\nret = %s\ndata=%s" \
                      % (labels_str, the_t.elapsed_ms(),ret_code, data))
            
            LOG.debug("Stored Labels %s in gm_id %s" % (labels_str, result_uid))

            self.server.select_folder(u'[Google Mail]/Drafts', readonly = self.readonly_folder) # go to current folder
        
            # check if it is ok otherwise exception
            if ret_code != 'OK':
                raise PushEmailError("Cannot add Labels %s to email with uid %d. Error:%s" % (labels_str, result_uid, data))
        
        return result_uid

def decode_labels(labels):
    """
       Decode labels when they are received as utf7 entities or numbers
    """
    new_labels = []
    for label in labels:
        if isinstance(label, (int, long, float, complex)):
            label = str(label) 
        new_labels.append(utf7_decode(label))

    return new_labels

# utf7 conversion functions
def utf7_encode(s): #pylint: disable=C0103
    """encode in utf7"""
    if isinstance(s, str) and sum(n for n in (ord(c) for c in s) if n > 127):
        raise ValueError("%r contains characters not valid in a str folder name. "
                              "Convert to unicode first?" % s)

    r = [] #pylint: disable=C0103
    _in = []
    for c in s: #pylint: disable=C0103
        if ord(c) in (range(0x20, 0x26) + range(0x27, 0x7f)):
            if _in:
                r.extend(['&', utf7_modified_base64(''.join(_in)), '-'])
                del _in[:]
            r.append(str(c))
        elif c == '&':
            if _in:
                r.extend(['&', utf7_modified_base64(''.join(_in)), '-'])
                del _in[:]
            r.append('&-')
        else:
            _in.append(c)
    if _in:
        r.extend(['&', utf7_modified_base64(''.join(_in)), '-'])
    return ''.join(r)


def utf7_decode(s): #pylint: disable=C0103
    """decode utf7"""
    r = [] #pylint: disable=C0103
    decode = []
    for c in s: #pylint: disable=C0103
        if c == '&' and not decode:
            decode.append('&')
        elif c == '-' and decode:
            if len(decode) == 1:
                r.append('&')
            else:
                r.append(utf7_modified_unbase64(''.join(decode[1:])))
            decode = []
        elif decode:
            decode.append(c)
        else:
            r.append(c)
    if decode:
        r.append(utf7_modified_unbase64(''.join(decode[1:])))
    out = ''.join(r)

    if not isinstance(out, unicode):
        out = unicode(out, 'latin-1')
    return out


def utf7_modified_base64(s): #pylint: disable=C0103
    """utf7 base64"""
    s_utf7 = s.encode('utf-7')
    return s_utf7[1:-1].replace('/', ',')


def utf7_modified_unbase64(s): #pylint: disable=C0103
    """ utf7 unbase64"""
    s_utf7 = '+' + s.replace(',', '/') + '-'
    return s_utf7.decode('utf-7')




# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contains the class monkey patching IMAPClient and imaplib

'''
import zlib
import time
import datetime
import re
import socket
import ssl
import cStringIO
import os

import imaplib  #for the exception
import imapclient

#enable imap debugging if GMV_IMAP_DEBUG is set 
if os.getenv("GMV_IMAP_DEBUG"):
    imaplib.Debug = 4 #enable debugging

#to enable imap debugging and see all command
#imaplib.Debug = 4 #enable debugging

INTERNALDATE_RE = re.compile(r'.*INTERNALDATE "'
r'(?P<day>[ 0123][0-9])-(?P<mon>[A-Z][a-z][a-z])-(?P<year>[0-9][0-9][0-9][0-9])'
r' (?P<hour>[0-9][0-9]):(?P<min>[0-9][0-9]):(?P<sec>[0-9][0-9])'
r' (?P<zonen>[-+])(?P<zoneh>[0-9][0-9])(?P<zonem>[0-9][0-9])'
r'"')

MON2NUM = {'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 'May': 5, 'Jun': 6,
        'Jul': 7, 'Aug': 8, 'Sep': 9, 'Oct': 10, 'Nov': 11, 'Dec': 12}

#need to monkey patch _convert_INTERNALDATE to work with imaplib2
#modification of IMAPClient
def mod_convert_INTERNALDATE(date_string, normalise_times=True):#pylint: disable=C0103
    """
       monkey patched convert_INTERNALDATE
    """
    mon = INTERNALDATE_RE.match('INTERNALDATE "%s"' % date_string)
    if not mon:
        raise ValueError("couldn't parse date %r" % date_string)
    
    zoneh = int(mon.group('zoneh'))
    zonem = (zoneh * 60) + int(mon.group('zonem'))
    if mon.group('zonen') == '-':
        zonem = -zonem
    timez = imapclient.fixed_offset.FixedOffset(zonem)
    
    year    = int(mon.group('year'))
    the_mon = MON2NUM[mon.group('mon')]
    day     = int(mon.group('day'))
    hour    = int(mon.group('hour'))
    minute  = int(mon.group('min'))
    sec = int(mon.group('sec'))
    
    the_dt = datetime.datetime(year, the_mon, day, hour, minute, sec, 0, timez)
    
    if normalise_times:
        # Normalise to host system's timezone
        return the_dt.astimezone(imapclient.fixed_offset.FixedOffset.for_system()).replace(tzinfo=None)
    return the_dt

#monkey patching is done here
imapclient.response_parser._convert_INTERNALDATE = mod_convert_INTERNALDATE

#monkey patching add compress in COMMANDS of imap
imaplib.Commands['COMPRESS'] = ('AUTH', 'SELECTED')

class IMAP4COMPSSL(imaplib.IMAP4_SSL): #pylint:disable=R0904
    """
       Add support for compression inspired by inspired by http://www.janeelix.com/piers/python/py2html.cgi/piers/python/imaplib2
    """
    SOCK_TIMEOUT = 70 # set a socket timeout of 70 sec to avoid for ever blockage in ssl.read

    def __init__(self, host = '', port = imaplib.IMAP4_SSL_PORT, keyfile = None, certfile = None):
        """
           constructor
        """
        self.compressor = None
        self.decompressor = None
        
        imaplib.IMAP4_SSL.__init__(self, host, port, keyfile, certfile)
        
    def activate_compression(self):
        """
           activate_compressing()
           Enable deflate compression on the socket (RFC 4978).
        """
        # rfc 1951 - pure DEFLATE, so use -15 for both windows
        self.decompressor = zlib.decompressobj(-15)
        self.compressor   = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, -15)
        
    def open(self, host = '', port = imaplib.IMAP4_SSL_PORT): 
        """Setup connection to remote server on "host:port".
           (default: localhost:standard IMAP4 SSL port).
           This connection will be used by the routines:
           read, readline, send, shutdown.
        """
        self.host   = host
        self.port   = port

        self.sock   = socket.create_connection((host, port), self.SOCK_TIMEOUT) #add so_timeout  

        #self.sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1) #try to set TCP NO DELAY to increase performances

        self.sslobj = ssl.wrap_socket(self.sock, self.keyfile, self.certfile)
        
        # This is the last correction added to avoid memory fragmentation in imaplib
        # makefile creates a file object that makes use of cStringIO to avoid mem fragmentation
        # it could be used without the compression 
        # (maybe make 2 set of methods without compression and with compression)
        #self.file   = self.sslobj.makefile('rb')
    
    def read(self, size):
        """
            Read 'size' bytes from remote.
            Call _intern_read that takes care of the compression
        """
        
        chunks = cStringIO.StringIO() #use cStringIO.cStringIO to avoir too much fragmentation
        read = 0
        while read < size:
            data = self._intern_read(min(size-read, 16384)) #never ask more than 16384 because imaplib can do it
            if not data: 
                #to avoid infinite looping due to empty string returned
                raise self.abort('Gmvault ssl socket error: EOF') 
            read += len(data)
            chunks.write(data)
        
        return chunks.getvalue() #return the cStringIO content
  
    def _intern_read(self, size):
        """
            Read at most 'size' bytes from remote.
            Takes care of the compression
        """
        if self.decompressor is None:
            return self.sslobj.read(size)

        if self.decompressor.unconsumed_tail:
            data = self.decompressor.unconsumed_tail
        else:
            data = self.sslobj.read(8192) #Fixed buffer size. maybe change to 16384

        return self.decompressor.decompress(data, size)
        
    def readline(self):
        """Read line from remote."""
        line = cStringIO.StringIO() #use cStringIO to avoid memory fragmentation
        while 1:
            #make use of read that takes care of the compression
            #it could be simplified without compression
            char = self.read(1) 
            line.write(char)
            if char in ("\n", ""): 
                return line.getvalue()
    
    def shutdown(self):
        """Close I/O established in "open"."""
        #self.file.close() #if file created
        self.sock.close()
        
      
    def send(self, data):
        """send(data)
        Send 'data' to remote."""
        if self.compressor is not None:
            data = self.compressor.compress(data)
            data += self.compressor.flush(zlib.Z_SYNC_FLUSH)
        self.sslobj.sendall(data)
       
def seq_to_parenlist(flags):
    """Convert a sequence of strings into parenthised list string for
    use with IMAP commands.
    """
    if isinstance(flags, str):
        flags = (flags,)
    elif not isinstance(flags, (tuple, list)):
        raise ValueError('invalid flags list: %r' % flags)
    return '(%s)' % ' '.join(flags)
    
class MonkeyIMAPClient(imapclient.IMAPClient): #pylint:disable=R0903,R0904
    """
       Need to extend the IMAPClient to do more things such as compression
       Compression inspired by http://www.janeelix.com/piers/python/py2html.cgi/piers/python/imaplib2
    """
    
    def __init__(self, host, port=None, use_uid=True, need_ssl=False):
        """
           constructor
        """
        super(MonkeyIMAPClient, self).__init__(host, port, use_uid, need_ssl)
    
    def _create_IMAP4(self): #pylint:disable=C0103
        """
           Factory method creating an IMAPCOMPSSL or a standard IMAP4 Class
        """
        imap_class = self.ssl and IMAP4COMPSSL or imaplib.IMAP4
        return imap_class(self.host, self.port)
    
    def xoauth_login(self, xoauth_cred ):
        """
           Connect with xoauth
           Redefine this method to suppress dependency to oauth2 (non-necessary)
        """

        typ, data = self._imap.authenticate('XOAUTH', lambda x: xoauth_cred)
        self._checkok('authenticate', typ, data)
        return data[0]  
    
    def search(self, criteria):
        """
           Perform a imap search or gmail search
        """
        if criteria.get('type','') == 'imap':
            #encoding criteria in utf-8
            req     = criteria['req'].encode('utf-8')
            charset = 'utf-8'
            return super(MonkeyIMAPClient, self).search(req, charset)
        elif criteria.get('type','') == 'gmail':
            return self.gmail_search(criteria.get('req',''))
        else:
            raise Exception("Unknown search type %s" % (criteria.get('type','no request type passed')))
        
    def gmail_search(self, criteria):
        """
           perform a search with gmailsearch criteria.
           eg, subject:Hello World
        """  
        criteria = criteria.replace('\\', '\\\\')
        criteria = criteria.replace('"', '\\"')

        #working but cannot send that understand when non ascii chars are used
        #args = ['CHARSET', 'utf-8', 'X-GM-RAW', '"%s"' % (criteria)]
        #typ, data = self._imap.uid('SEARCH', *args)

        #working Literal search 
        self._imap.literal = '"%s"' % (criteria)
        self._imap.literal = imaplib.MapCRLF.sub(imaplib.CRLF, self._imap.literal)
        self._imap.literal = self._imap.literal.encode("utf-8")
 
        #use uid to keep the imap ids consistent
        args = ['CHARSET', 'utf-8', 'X-GM-RAW']
        typ, data = self._imap.uid('SEARCH', *args) #pylint: disable=W0142
        
        self._checkok('search', typ, data)
        if data == [None]: # no untagged responses...
            return [ ]

        return [ long(i) for i in data[0].split() ]
    
    def append(self, folder, msg, flags=(), msg_time=None):
        """Append a message to *folder*.

        *msg* should be a string contains the full message including
        headers.

        *flags* should be a sequence of message flags to set. If not
        specified no flags will be set.

        *msg_time* is an optional datetime instance specifying the
        date and time to set on the message. The server will set a
        time if it isn't specified. If *msg_time* contains timezone
        information (tzinfo), this will be honoured. Otherwise the
        local machine's time zone sent to the server.

        Returns the APPEND response as returned by the server.
        """
        if msg_time:
            time_val = time.mktime(msg_time.timetuple())
        else:
            time_val = None

        flags_list = seq_to_parenlist(flags)

        typ, data = self._imap.append(self._encode_folder_name(folder) if folder else None,
                                      flags_list, time_val, msg)
        self._checkok('append', typ, data)

        return data[0]
    
    def enable_compression(self):
        """
        enable_compression()
        Ask the server to start compressing the connection.
        Should be called from user of this class after instantiation, as in:
            if 'COMPRESS=DEFLATE' in imapobj.capabilities:
                imapobj.enable_compression()
        """
        ret_code, _ = self._imap._simple_command('COMPRESS', 'DEFLATE') #pylint: disable=W0212
        if ret_code == 'OK':
            self._imap.activate_compression()
        else:
            #no errors for the moment
            pass

        



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import unittest
import base64

import gmv.gmvault as gmvault
import gmv.gmvault_utils as gmvault_utils
import gmv.imap_utils as imap_utils

def obfuscate_string(a_str):
    """ use base64 to obfuscate a string """
    return base64.b64encode(a_str)

def deobfuscate_string(a_str):
    """ deobfuscate a string """
    return base64.b64decode(a_str)

def read_password_file(a_path):
    """
       Read log:pass from a file in my home
    """
    pass_file = open(a_path)
    line = pass_file.readline()
    (login, passwd) = line.split(":")
    
    return (deobfuscate_string(login.strip()), deobfuscate_string(passwd.strip()))

def delete_db_dir(a_db_dir):
    """
       delete the db directory
    """
    gmvault_utils.delete_all_under(a_db_dir, delete_top_dir = True)


class TestGMVaultValidation(unittest.TestCase): #pylint:disable=R0904
    """
       Validation Tests
    """

    def __init__(self, stuff):
        """ constructor """
        super(TestGMVaultValidation, self).__init__(stuff)
        
        self.test_login  = None
        self.test_passwd = None 
        
        self.default_dir = "/tmp/gmvault-tests"
    
    def setUp(self): #pylint:disable=C0103
        self.test_login, self.test_passwd = read_password_file('/homespace/gaubert/.ssh/gsync_passwd')
                
    def test_help_msg_spawned_by_def(self):
        """
           spawn python gmv_runner account > help_msg_spawned.txt
           check that res is 0 or 1
        """
        credential  = { 'type' : 'passwd', 'value': self.test_passwd}
        test_db_dir = "/tmp/gmvault-tests"
        
        restorer = gmvault.GMVaulter(test_db_dir, 'imap.gmail.com', 993, self.test_login, credential, \
                                     read_only_access = False)
        
        restorer.restore() #restore all emails from this essential-db
        
        #need to check that all labels are there for emails in essential
        gmail_ids = restorer.gstorer.get_all_existing_gmail_ids()
        
        for gm_id in gmail_ids:
            #get disk_metadata
            disk_metadata   = restorer.gstorer.unbury_metadata(gm_id)
            
            # get online_metadata 
            online_metadata = restorer.src.fetch(gm_id, imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA) 
            
            #compare metadata
            for key in disk_metadata:
                self.assertEquals(disk_metadata[key], online_metadata[key])
            

def tests():
    """
       main test function
    """
    suite = unittest.TestLoader().loadTestsFromTestCase(TestGMVaultValidation)
    unittest.TextTestRunner(verbosity=2).run(suite)
 
if __name__ == '__main__':
    
    tests()
